/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2004 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include <assert.h>

#include <SDL.h>

#include <GL/gl.h>

#include "engine.h"

/*
 * Init player commands.
 */
int engine_init_cmds(Engine *engine) {
  /* Disable some events */
  SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
  SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
  SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
  SDL_EventState(SDL_MOUSEBUTTONUP, SDL_IGNORE);
  SDL_EventState(SDL_JOYAXISMOTION, SDL_IGNORE);
  SDL_EventState(SDL_JOYBALLMOTION, SDL_IGNORE);
  SDL_EventState(SDL_JOYHATMOTION, SDL_IGNORE);
  SDL_EventState(SDL_JOYBUTTONDOWN, SDL_IGNORE);
  SDL_EventState(SDL_JOYBUTTONUP, SDL_IGNORE);
  SDL_EventState(SDL_SYSWMEVENT, SDL_IGNORE);
  SDL_EventState(SDL_VIDEORESIZE, SDL_IGNORE);
  SDL_EventState(SDL_USEREVENT, SDL_IGNORE);

  return 0;
}

/*
 * Get player commands.
 */
int engine_get_cmds(Engine *engine) {
  short done = 0;
  SpaceShip *player_ship = NULL;
  SDL_Event event;
  static short kctrl = 0;
  static short kshift = 0;
  static short roll = 1; /* roll?yaw */

  /* Assertion */
  assert(engine);
  assert(engine->scene);

  player_ship = scene_player_ship(engine->scene);

  while(SDL_PollEvent(&event)) {
    switch(event.type) {

    case SDL_QUIT:
      done = 1;
      break;

    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case SDLK_ESCAPE:
	done = 1; /* quit */
	break;
      case SDLK_LCTRL:
      case SDLK_RCTRL:
	kctrl = 1;
	break;
      case SDLK_LSHIFT:
      case SDLK_RSHIFT:
	kshift = 1;
	break;
      case SDLK_q:
	/* Accelerate */
	space_ship_accelerate(player_ship);
	break;
      case SDLK_w:
	/* Decelerate */
	space_ship_decelerate(player_ship);
	break;
      case SDLK_UP:
	if(kshift) {
	  /* Move Up */
	  space_ship_move_up(player_ship);
	}
	else {
	  /* Pitch Down */
	  space_ship_pitch_down(player_ship);
	}
	break;
      case SDLK_DOWN:
	if(kshift) {
	  /* Move Down */
	  space_ship_move_down(player_ship);
	}
	else {
	  /* Pitch Up */
	  space_ship_pitch_up(player_ship);
	}
	break;
      case SDLK_LEFT:
	if(kshift) {
	  /* Move Left */
	  space_ship_move_left(player_ship);
	}
	else {
	  if(roll) {
	    /* Roll Left */
	    space_ship_roll_left(player_ship);
	  }
	  else {
	    /* Yaw Left */
	    space_ship_yaw_left(player_ship);
	  }
	}
	break;
      case SDLK_RIGHT:
	if(kshift) {
	  /* Move Right */
	  space_ship_move_right(player_ship);
	}
	else {
	  if(roll) {
	    /* Roll Right */
	    space_ship_roll_right(player_ship);
	  }
	  else {
	    /* Yaw Right */
	    space_ship_yaw_right(player_ship);
	  }
	}
	break;
      case SDLK_y:
	/* Swap Roll/Yaw */
	roll = 1 - roll;
	break;
      case SDLK_m:
	/* Swap Free/Assisted flight */
	space_ship_swap_assisted_mode(player_ship);
	break;
      case SDLK_F5:
	/* Stop (auto) */
	space_ship_to_break_off(player_ship);
	break;
      case SDLK_F6:
	/* Approach (auto) */
	space_ship_to_come_near(player_ship);
	break;
      case SDLK_F7:
	/* Formation (auto) */
	space_ship_in_formation(player_ship);
	break;
      case SDLK_F8:
	/* Moor (auto) */
	space_ship_to_moor(player_ship);
	break;
      case SDLK_F1:
	/* Console NAV */
	space_ship_vdu_nav(player_ship);
	break;
      case SDLK_F2:
	/* Console ARM */
	space_ship_vdu_arm(player_ship);
	break;
      case SDLK_F3:
	/* Console NRJ */
	space_ship_vdu_nrj(player_ship);
	break;
      case SDLK_F4:
	/* Console CMD */
	space_ship_vdu_cmd(player_ship);
	break;
      case SDLK_h:
	/* Change HUD details */
	space_ship_change_hud(player_ship);
	break;
      case SDLK_n:
	if(kctrl) {
	  /* Normals (two sides) */
	  glFrontFace(GL_CCW);
	}
	break;
      case SDLK_c:
	if(kctrl) {
	  /* Cull faces (on/off) */
	  if(glIsEnabled(GL_CULL_FACE)) {
	    glDisable(GL_CULL_FACE);
	  }
	  else {
	    glEnable(GL_CULL_FACE);
	  }
	}
	break;
      case SDLK_d:
	if(kctrl) {
	  /* Depth buffer (on/off) */
	  if(glIsEnabled(GL_DEPTH_TEST)) {
	    glDisable(GL_DEPTH_TEST);
	  }
	  else {
	    glEnable(GL_DEPTH_TEST);
	  }
	}
	break;
      case SDLK_s:
	if(kctrl) {
	  /* Shade model */
	  glShadeModel(GL_SMOOTH);
	  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	break;
      default:
	break;
      }
      break;

    case SDL_KEYUP:
      switch(event.key.keysym.sym) {
      case SDLK_LCTRL:
      case SDLK_RCTRL:
	kctrl = 0;
	break;
      case SDLK_LSHIFT:
      case SDLK_RSHIFT:
	kshift = 0;
	break;
      case SDLK_q:
      case SDLK_w:
	/* Stop accelerating/decelerating */
	space_ship_stop_move_forward(player_ship);
	break;
      case SDLK_UP:
      case SDLK_DOWN:
	/* Stop moving up/down */
	space_ship_stop_move_up_down(player_ship);
	/* Stop pitching */
	space_ship_stop_pitch(player_ship);
	break;
      case SDLK_LEFT:
      case SDLK_RIGHT:
	/* Stop moving left/right */
	space_ship_stop_move_lateral(player_ship);
	/* Stop rolling */
	space_ship_stop_roll(player_ship);
	/* Stop heading */
	space_ship_stop_yaw(player_ship);
	break;
      default:
	break;
      }
      break;

    default:
      break;
    }
  } /* end while */

  return done;
}

