/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include <assert.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/*#include <sys/types.h>*/
#ifdef UNISTD
#include <unistd.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include "getopt.h" /* part of GNU C library */

#include <libxml/tree.h>
#include <libxml/xpath.h>

#include "sdlib.h"

#include "engine.h"

/*
 * Print help/usage.
 */
static int show_help(char *progname) {
  printline("\nUsage: %s [options]...\n", progname);
  printline("Options:\n"
	    "  --help                  print this message\n"
	    "  --verbose               display some infos\n"
	    "  --atest                 test mode\n"
	    "  --geometry 800x600      resolution\n"
	    "  --full                  fullscreen\n"
	    "  --conf file.xml         read a config file\n"
	    "  --font file.txf         use a particular font\n"
	    "  --mission file.xml      load a mission\n");
  printline("Keys:\n"
	    "  <Esc>          quit\n"
	    "Controls:\n"
	    "  <Up/Down>      pitch up/down\n"
	    "  <Left/Right>   roll or yaw left/right\n"
	    "  <y>            swap roll/yaw axis\n"
	    "  <q/w>          main prop.\n"
	    "  <+/->          main prop. (accelerate/decelerate)\n"
	    "  Shift + <Dir>  lateral prop.\n"
	    "  <m>            assisted mode (on/off)\n\n");
  return 0;
}

/*
 * Search a key (from path) in xml file;
 * Return the value.
 */
static char *search_cfg_key(xmlDocPtr xml_doc, const char *key) {
  const char *path_attribute = "/Stardust/configuration/item[@name='%s']/text()";
  char *path;
  char *value = NULL;
  xmlXPathContextPtr xml_context = NULL;
  xmlXPathObjectPtr xml_obj;

  /* Assertion */
  assert(key);
  assert(xml_doc);

  xmlXPathInit();
  xml_context = xmlXPathNewContext(xml_doc);

  /* Find key (path) */
  path = malloc((strlen(path_attribute) + strlen(key)) * sizeof(char));
  if(!path) {
    sd_error("Not enough memory!");
  }
  sprintf(path, path_attribute, key);

  xml_obj = xmlXPathEval(path, xml_context);
  free(path);

  if( (xml_obj->type == XPATH_NODESET) &&
      (xml_obj->nodesetval) ) {
    if(xml_obj->nodesetval->nodeNr) {
      xmlNodePtr node = xml_obj->nodesetval->nodeTab[0];
      if( (node->type == XML_TEXT_NODE) ||
	  (node->type == XML_CDATA_SECTION_NODE) ) {
	value = strdup(node->content);
      }
    }
  }

  /* Free memory */
  xmlXPathFreeObject(xml_obj);
  xmlXPathFreeContext(xml_context);

  return value;
}

/*
 * Read a configuration file (in XML).
 */
static int engine_read_conf(Engine *engine, const char *conf_file) {
  const char *keys[] = { "none", 
			 "atest", 
			 "verbose", 
			 "fps", 
			 "font", 
			 "geometry", 
			 "fullscreen", NULL };
  unsigned int k = 0;
  xmlDocPtr xml_doc = NULL;
  char *val = NULL;

  /* Assertion */
  assert(engine);
  assert(conf_file);

  /* Parse xml file */
  xmlInitParser();
  xml_doc = xmlParseFile(conf_file);
  if(!xml_doc) {
    sd_error("Could not parse the document %s!", conf_file);
  }

  while(keys[k]) {
    if(!strcmp(keys[k], "none")) {
      printline("Reading config file: %s\n", conf_file);
      ifdebug( xmlDocDump(stderr, xml_doc); );
    }
    else if(!strcmp(keys[k], "atest")) {
      if((val = search_cfg_key(xml_doc, keys[k]))) {
	dprintf("Test mode: %s\n", val);
	engine->atest = (!strcmp(val, "on"))? 1:0;
      }
    }
    else if(!strcmp(keys[k], "verbose")) {
      if((val = search_cfg_key(xml_doc, keys[k]))) {
	dprintf("Verbose mode: %s\n", val);
	engine->verbose = (!strcmp(val, "on"))? 1:0;
      }
    }
    else if(!strcmp(keys[k], "fps")) {
      if((val = search_cfg_key(xml_doc, keys[k]))) {
	dprintf("FPS requested: %s\n", val);
	engine->fps = atoi(val);
      }
    }
    else if(!strcmp(keys[k], "font")) {
      if((val = search_cfg_key(xml_doc, keys[k]))) {
	dprintf("Tex font: %s\n", val);
	sscanf(val, "%s", engine->txf_file);
      }
    }
    else if(!strcmp(keys[k], "geometry")) {
      if((val = search_cfg_key(xml_doc, keys[k]))) {
	dprintf("Geometry requested: %s\n", val);
	sscanf(val, "%dx%d", &(engine->width), &(engine->height));
        engine->aspect = (float)(engine->width)/(float)(engine->height);
      }
    }
    else if(!strcmp(keys[k], "fullscreen")) {
      if((val = search_cfg_key(xml_doc, keys[k]))) {
	dprintf("Fullscreen mode: %s\n", val);
	if(!strcmp(val, "on")) {
	  engine->flags |= SDL_FULLSCREEN;
	}
      }
    }
    else {
      sd_debug("Must never happen!");
    }
    k++;
  }
  dprintf("\n");

  /* Free memory */
  xmlFreeDoc(xml_doc);

  return 0;
}

/*
 * Init engine (conf).
 */
int engine_init(Engine *engine, int argc, char **argv) {
  /* Assertion */
  assert(engine);

  /* Engine status */
  engine->atest   = 0;
  engine->verbose = 0;

  engine->session = SESSION_LOCAL;

  /* Initialize  windows properties */
  /*engine->screen = NULL;*/
  engine->flags = SDL_OPENGL; /* | SDL_RESIZABLE | SDL_FULLSCREEN; */
  engine->width  = 800;
  engine->height = 600;
  engine->bpp    = 16;

  /* Initialize camera */
  engine->fovy  = 80.0;
  engine->aspect = 4.0/3.0;
  engine->near = 2.0;
  engine->far  = 100000.0;

  /* Frame rate */
  engine->fps = 0;
  engine->frame_rate = 0; 
  engine->ticks        = 0;
  engine->elapsed_time = 0;

  sd_debug("Prefix directory: %s\n", PREFIX);
  sd_debug("Data directory: %s\n\n", DATA_DIR);

#ifdef DEBUG /* override atest in debug mode */
  engine->atest = 1;
#endif

  /* Initialize world/mission */
  sprintf(engine->mission_file, "data/xml/scenes/fleet_iwar.xml"); /* TODO: remove */

  /* Initialize hud font */
  sprintf(engine->txf_file, "data/txf/helvetica.txf");
  
  assert(argc); /* assertion */
  {
    int c, opt_ind=0; /* getopt parameters */
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"verbose", 0, 0, 'v'},
      {"atest", 0, 0, 't'},
      {"geometry", 1, 0, 'g'},
      {"full", 0, 0, 'f'},
      {"fps", 1, 0, 'r'},
      {"conf", 1, 0, 'c'},
      {"font", 1, 0, 'a'},
      {"mission", 1, 0, 'm'},
      {0, 0, 0, 0}
    };
    
    /* Filter command line options */
    while( (c=getopt_long_only(argc, argv, "hvtg:fr:c:a:m:", 
			       long_options, &opt_ind)) != EOF ) {
      switch(c) {
      case 'h':
	show_help(argv[0]);
	exit(0);
      case 'v':
	printline("Verbose mode activated!\n");
	engine->verbose = 1;
	break;
      case 't':
	printline("Test mode activated!\n");
	engine->atest = 1;
	break;
      case 'g':
	sscanf(optarg, "%dx%d", &(engine->width), &(engine->height));
        engine->aspect = (float)(engine->width)/(float)(engine->height);
	printline("Display geometry requested: %dx%d (%.2f)\n", 
		  engine->width, engine->height, engine->aspect);
	break;
      case 'f':
	engine->flags |= SDL_FULLSCREEN;
	break;
      case 'r':
	printline("Frame rate requested: %d FPS\n", atoi(optarg));
	engine->fps = atoi(optarg);
	break;
      case 'c':
	engine_read_conf(engine, optarg);
	break;
      case 'a':
	sscanf(optarg, "%s", engine->txf_file);
	printline("Tex font: %s\n", engine->txf_file);
	break;
      case 'm':
	sscanf(optarg, "%s", engine->mission_file);
	printline("Load mission: %s\n", engine->mission_file);
	break;
      default:
	break;
      }
    }
    printline("\n");
    /* Other args */
    if(opt_ind < argc) {
    }
    else { /* too few args */
    }
  }

  return 0;
}

