/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include <assert.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <SDL.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "sdlib.h"

#include "engine.h"

/* 
 * Check some video modes.
 */
static int check_sdl_video_modes(void) {
  SDL_VideoInfo *video_info;
  SDL_Rect **video_modes;
  SDL_PixelFormat *best_vfmt = NULL;
  char driver_name[8];
  Uint8 bpp;
  int i;

  /* Get some infos on system */
  if(SDL_VideoDriverName(driver_name, 8)) {
     printline("VideoDriverName: %s\n", driver_name);
  }

  video_info = (SDL_VideoInfo *)SDL_GetVideoInfo();
  if(video_info) {
    printline("  hardware surfaces? %d\n", video_info->hw_available);
    printline("  window manager? %d\n", video_info->wm_available);
    printline("  hardware accel? %d\n", video_info->blit_hw);
    printline("  video memory: %u Kb\n", video_info->video_mem);
    best_vfmt = video_info->vfmt;
    printline("'best' available mode: %d bpp\n", best_vfmt->BitsPerPixel);
  }

  /* Get available software/hardware/fullscreen modes */
  video_modes = SDL_ListModes(NULL,
			      SDL_OPENGL |
			      /*SDL_RESIZABLE |*/
                              SDL_FULLSCREEN);
  /* Check is there are any modes available */
  if(video_modes == (SDL_Rect **)0) {
    sd_error("No video modes available!");
  }
  /* Check if or resolution is restricted */
  if(video_modes == (SDL_Rect **)-1) {
    printline("All resolutions available.\n");
  }
  else { /* print valid modes */
    printline("Available video modes:\n");
    for(i=0; video_modes[i]; ++i) {
      bpp = SDL_VideoModeOK(video_modes[i]->w, video_modes[i]->h, 
			    best_vfmt->BitsPerPixel, 
			    SDL_OPENGL |
			    /*SDL_RESIZABLE |*/
			    SDL_FULLSCREEN);
      printline("  %d x %d @ %d bpp recommended.\n", 
		video_modes[i]->w, video_modes[i]->h, bpp);
    }
  }

  printline("\n");
  return 0;
}

/* 
 * Print some GL infos. 
 */
static int print_gl_infos(void) {
  printline("GL_VENDOR: %s\n", (char *)glGetString(GL_VENDOR));
  printline("GL_VERSION: %s\n", (char *)glGetString(GL_VERSION));
  printline("GL_RENDERER: %s\n", (char *)glGetString(GL_RENDERER));
  printline("GL_EXTENSIONS: %s\n", (char *)glGetString(GL_EXTENSIONS));
  printline("GLU_VERSION: %s\n", (char *)gluGetString(GLU_VERSION));
  printline("GLU_EXTENSIONS: %s\n", (char *)gluGetString(GLU_EXTENSIONS));
  printline("\n");
  return 0;
}

/*
 * Initialise graphics.
 */
int engine_init_graphics(Engine *engine) {
  SDL_Surface *screen;

  /* Initialize the SDL library */
  if(SDL_Init(SDL_INIT_VIDEO 
	      | SDL_INIT_EVENTTHREAD 
	      | SDL_INIT_NOPARACHUTE) < 0) {
    sd_error("Couldn't initialize SDL: %s!", SDL_GetError());
  }
  atexit(SDL_Quit);

  /* Assertion */
  assert(engine);

  if(engine->verbose) check_sdl_video_modes(); /* check some video modes */

  { /* Let's get some video information */
    const SDL_VideoInfo *info = SDL_GetVideoInfo();
    engine->bpp = info->vfmt->BitsPerPixel; /* 'best' available mode */
  }

  /* Now, we want to setup our OpenGL window */
  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 6);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  /* Initialize video, have a preference for 16-bit, but accept any depth */
  screen = SDL_SetVideoMode(engine->width, engine->height, 
			    engine->bpp, engine->flags);
  if(!screen) {
    sd_error("Couldn't initialize %d x %d @ %d bpp: %s!", 
	     engine->width, engine->height, engine->bpp, SDL_GetError());
  }

  engine->width = screen->w;
  engine->height = screen->h;
  engine->bpp = screen->format->BitsPerPixel;
  printline("SDL screen: %d x %d @ %d bpp\n\n", 
	  screen->w, screen->h, screen->format->BitsPerPixel);

  SDL_WM_SetCaption("Stardust "VERSION, "Stardust");

  if(engine->verbose) print_gl_infos(); /* display some OpenGL infos */

  { /* Init some OpenGL state */
    GLfloat global_ambient[4] = {0.2, 0.2, 0.2, 1.0};
    GLfloat light0_pos[4] = {0.0, 0.0, 1.0, 0.0}; /* directional */
    GLfloat light0_ambient[4] = {0.2, 0.2, 0.2, 1.0};
    GLfloat light0_diffuse[4] = {1.0, 1.0, 1.0, 1.0};
    GLfloat light0_specular[4] = {1.0, 1.0, 1.0, 1.0};
    GLfloat mat_specular[4] = {1.0, 1.0, 1.0, 1.0};
    GLfloat mat_emission[4] = {0.0, 0.0, 0.0, 1.0};
    GLint mat_shininess = 128;
     
    /* Lighting */
    glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
    glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);

    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_EMISSION, mat_emission);
    glMateriali(GL_FRONT, GL_SHININESS, mat_shininess);

    /* Texture mapping */
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glEnable(GL_TEXTURE_2D);
#ifdef GL_EXT_polygon_offset
    glEnable(GL_POLYGON_OFFSET_EXT);
    glPolygonOffset(0.0, -3);
#endif
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    /* Antialiasing */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_DONT_CARE);
    glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
    glHint(GL_POINT_SMOOTH_HINT, GL_DONT_CARE);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_DONT_CARE);
    glEnable(GL_PERSPECTIVE_CORRECTION_HINT);
    glEnable(GL_POLYGON_SMOOTH_HINT);

    /* Z-buffer, back-face culling */
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    
    glDisable(GL_DITHER);
    glShadeModel(GL_SMOOTH);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
     
    glClearColor(0.0, 0.0, 0.0, 1.0); 

    /* Initialize projection and modelview matrix */
    glViewport(0, 0, engine->width, engine->height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(engine->fovy, engine->aspect, engine->near, engine->far);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  }

  return 0;
}
