/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2004 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include <assert.h>

#include "sdlib.h"

#include "hud.h"

/*
 * Constructor.
 */
Hud *hud_new(void) {
  /* Allocation de la structure */
  Hud *hud = (Hud *)malloc(sizeof(Hud));
  if(!hud) {
    sd_error("Can't allocate memory (Hud)!");
  }

  /* Hud init */
  hud->width  = 0.0;
  hud->height = 0.0;
  hud->dz     = 0.0;
  hud->unit   = 0.0;
  hud->ratio  = 0.0;
  
  hud->w2  = 0.0;
  hud->w3  = 0.0;
  hud->w4  = 0.0;
  hud->w5  = 0.0;
  hud->w8  = 0.0;
  hud->w10 = 0.0;
  hud->w20 = 0.0;
  hud->w30 = 0.0;
  hud->w40 = 0.0;
  hud->w80 = 0.0;

  hud->h2  = 0.0;
  hud->h5  = 0.0;
  
  hud->font = font_new();

  hud->intest = 0;

  return hud;
}

/*
 * Destructor.
 */
int hud_destroy(Hud *hud) {
  /* Assertion */
  assert(hud);
  assert(hud->font);

  /* Free memory */
  font_destroy(hud->font);
  free(hud);

  return 0;
}

/*
 * Init hud (width, height, pos, font).
 */
int hud_init(Hud *hud, float w, float h, float z, const char *txf_file, const short atest) {
  /* Assertion */
  assert(hud);
  assert(hud->font);

  /* Initialization */
  hud->width  = w;
  hud->height = h;
  hud->dz     = z;
  hud->unit   = w / 100.0;
  hud->ratio  = w / h;

  hud->w2  = w / 2.0;
  hud->w3  = w / 3.0;
  hud->w4  = w / 4.0;
  hud->w5  = w / 5.0;
  hud->w8  = w / 8.0;
  hud->w10 = w / 10.0;
  hud->w20 = w / 20.0;
  hud->w30 = w / 30.0;
  hud->w40 = w / 40.0;
  hud->w80 = w / 80.0;

  hud->h2  = h / 2.0;
  hud->h5  = h / 5.0;

  font_init(hud->font, txf_file);

  hud->intest = atest;

  return 0;
}

/*
 * Draw hud (linked to player ship).
 */
int hud_draw(Hud *hud, SpaceShip *player_ship, const float frame_rate) {
  /* Assertion */
  assert(hud);

  glPushMatrix();
#if 1
  /* Change to orthographic projection */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-hud->w2, hud->w2, 
	  -hud->h2, hud->h2,
	  hud->dz, hud->dz + 1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
#else
  glLoadIdentity();
  glScalef(1.3, 1.3, 1.0);
#endif

  /* Change some OpenGL state */
  glDisable(GL_LIGHTING);

  glDepthMask(GL_FALSE);
  glDisable(GL_CULL_FACE);

  glDisable(GL_TEXTURE_2D);

  /* Disable line smooth */  
  glDisable(GL_LINE_SMOOTH);
  glDisable(GL_POINT_SMOOTH);

  /* Disable blending */
  glDisable(GL_BLEND);

  /* Disable transparency */
  glDisable(GL_ALPHA_TEST);

  switch(space_ship_hud_detail(player_ship)) {
  case HUD_HIGH_DETAIL:
    /* Enable line smooth */  
    glLineWidth(1.1);
    glEnable(GL_LINE_SMOOTH);
    glPointSize(1.1);
    glEnable(GL_POINT_SMOOTH);
  case HUD_MED_DETAIL:
    /* Enable blending */
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
  case HUD_LOW_DETAIL:
    /* Enable transparency */
    glAlphaFunc(GL_GREATER, 0.1f);
    glEnable(GL_ALPHA_TEST);
  case HUD_NO_DETAIL:
    /* Do nothing */
  default:
    /* Do nothing */
    break;
  }
  
  switch(space_ship_vdu(player_ship)) {
  case CONSOLE_NAV:
    /* Draw screen */
    hud_draw_screen(hud, player_ship);
    hud_draw_speed(hud, player_ship);
    hud_draw_flying_mode(hud, player_ship);
    hud_draw_warnings(hud, player_ship);

    /* Display messages */
    hud_draw_message(hud, player_ship, frame_rate);

    /* Draw orbe */
    hud_draw_orbe(hud, player_ship);

    /* Draw TRI */
    hud_draw_tri(hud, player_ship);

    /* Display player vessel infos */
    hud_draw_vessel(hud, player_ship);
    hud_draw_damages(hud, player_ship);

    /* Display contact infos */
    hud_draw_contact(hud, player_ship);
    hud_draw_camera(hud, player_ship);
    hud_draw_target(hud, player_ship);

    /* Draw NAV controls */
    hud_draw_nav(hud, player_ship);
    
    break;
  case CONSOLE_ARM:
    /* Draw screen */
    hud_draw_screen(hud, player_ship);
    hud_draw_speed(hud, player_ship);
    hud_draw_flying_mode(hud, player_ship);
    hud_draw_warnings(hud, player_ship);

    /* Display messages */
    hud_draw_message(hud, player_ship, frame_rate);

    /* Draw orbe */
    hud_draw_orbe(hud, player_ship);

    /* Draw TRI */
    hud_draw_tri(hud, player_ship);

    /* Display player vessel infos */
    hud_draw_vessel(hud, player_ship);
    hud_draw_resources(hud, player_ship);
    hud_draw_damages(hud, player_ship);

    /* Display contact infos */
    hud_draw_contacts(hud, player_ship);
    hud_draw_camera(hud, player_ship);
    hud_draw_target(hud, player_ship);

    /* Draw ARM controls */
    hud_draw_arm(hud, player_ship);
    
    break;
  case CONSOLE_NRJ:
    /* Draw screen */
    hud_draw_screen(hud, player_ship);

    /* Display messages */
    hud_draw_message(hud, player_ship, frame_rate);
    
    /* Draw TRI */
    hud_draw_tri(hud, player_ship);

    /* Display player vessel infos */
    hud_draw_vessel(hud, player_ship);
    hud_draw_resources(hud, player_ship);
    hud_draw_damages(hud, player_ship);

    /* Draw NRJ controls */
    hud_draw_nrj(hud, player_ship);

    break;
  case CONSOLE_CMD:
    /* Draw screen */
    hud_draw_screen(hud, player_ship);

    /* Display messages */
    hud_draw_message(hud, player_ship, frame_rate);

    /* Draw CMD controls */
    hud_draw_cmd(hud, player_ship);

    break;
  default:
    sd_warning("Must never happen!");
  }
  
  glPopMatrix();

  /* Re-init OpenGL */
  glDepthMask(GL_TRUE);
  glEnable(GL_CULL_FACE);
  
  glEnable(GL_LIGHTING);
  
  return 0;
}

