/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2004 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include <assert.h>

#include <math.h>

#include "sdlib.h"

#include "hud.h"

/*
 * Draw corners.
 */
static int hud_draw_screen_corners(Hud *hud) {
  float x, y, z, d;

  /* Assertion */
  assert(hud);

  /* Corner up-right */
  x = hud->w2 - hud->unit;
  y = hud->h2 - hud->unit;
  z = -hud->dz - 0.1;
  /* Size */
  d = hud->w40;

  /* Draw corners */
  glColor4f(1.0, 0.0, 0.0, 0.8);
  glBegin(GL_LINE_STRIP);
  glVertex3f(-x, -y+d, z);
  glVertex3f(-x, -y, z);
  glVertex3f(-x+d, -y, z);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex3f(x, -y+d, z);
  glVertex3f(x, -y, z);
  glVertex3f(x-d, -y, z);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex3f(x, y-d, z);
  glVertex3f(x, y, z);
  glVertex3f(x-d, y, z);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex3f(-x, y-d, z);
  glVertex3f(-x, y, z);
  glVertex3f(-x+d, y, z);
  glEnd();

  return 0;
}

/*
 * Draw console.
 */
static int hud_draw_console(Hud *hud, SpaceShip *player_ship) {
  char str[64];
  float x, y, z, d;

  /* Assertion */
  assert(hud);
  assert(player_ship);

  /* Edge middle-up */
  x = 0.0;
  y = hud->h2 - hud->unit;
  z = -hud->dz - 0.1;
  /* Size */
  d = 2.0*hud->unit;

  glColor4f(0.8, 0.8, 0.8, 0.8);

  /* Console mode */
  switch(space_ship_vdu(player_ship)) {
  case CONSOLE_NAV:
    sprintf(str, "NAV");
    break;
  case CONSOLE_ARM:
    sprintf(str, "ARM");
    break;
  case CONSOLE_NRJ:
    sprintf(str, "NRJ");
    break;
  case CONSOLE_CMD:
    sprintf(str, "CMD");
    break;
  default:
    sd_warning("Must never happen!");
  }
  font_draw(hud->font, str, 8, x-8.0*d, y-d, z);

  /* Camera mode */
  switch(space_ship_view(player_ship)) {
  case OUTER_VIEW:
    sprintf(str, "OUTER VIEW");
    break;
  case REAR_VIEW:
    sprintf(str, "REAR VIEW");
    break;
  case COCKPIT_VIEW:
    sprintf(str, "COCKPIT VIEW");
    break;
  default:
    sd_warning("Must never happen!");
  }
  font_draw(hud->font, str, 8, x-3.0*d, y-d, z);

  /* Hud mode */
  switch(space_ship_hud_detail(player_ship)) {
  case HUD_NO_DETAIL:
    sprintf(str, "NO HUD");
    break;
  case HUD_LOW_DETAIL:
    sprintf(str, "HUD LOW");
    break;
  case HUD_MED_DETAIL:
    sprintf(str, "HUD MED");
    break;
  case HUD_HIGH_DETAIL:
    sprintf(str, "HUD HIGH");
    break;
  default:
    sd_warning("Must never happen!");
  }
  font_draw(hud->font, str, 8, x+4.0*d, y-d, z);

  return 0;
}

/*
 * Draw screen.
 */
int hud_draw_screen(Hud *hud, SpaceShip *player_ship) {
  /* Assertion */
  assert(hud);
  assert(player_ship);

  /* Draw corners */
  hud_draw_screen_corners(hud);

  /* Draw console */
  hud_draw_console(hud, player_ship);

  return 0;
}

/*
 * Draw speed (edge middle-left).
 */
int hud_draw_speed(Hud *hud, SpaceShip *player_ship) {
  char str[64];
  float x, y, z, a, b;

  /* Assertion */
  assert(hud);
  assert(hud->font);
  assert(player_ship);

  /* Edge middle-left */
  x = -hud->w2 + 2.0*hud->unit; 
  y = 0.0;
  z = -hud->dz - 0.1;
  /* Text */
  a = x + hud->unit;
  b = y + hud->unit; /* line offset */

  /* Display player vessel speed */
  glColor4f(1.0, 1.0, 0.0, 0.8);
  sprintf(str, "Spd: %.0f m/s", 
	  floor(space_ship_speed(player_ship)/100)*100);
  font_draw(hud->font, str, 8, a, b, z);

  return 0;
}

/*
 * Display flying_mode (edge middle-up).
 */
int hud_draw_flying_mode(Hud *hud, SpaceShip *player_ship) {
  float x, y, z, d;

  /* Assertion */
  assert(hud);
  assert(hud->font);
  assert(player_ship);

  /* Edge middle-up */
  x = 0.0;
  y = hud->h2 - hud->unit;
  z = -hud->dz - 0.1;
  /* Size */
  d = 2.0*hud->unit;

  /* Display flying_mode */
  glColor4f(1.0, 1.0, 0.0, 0.8);
  font_draw(hud->font, "S L H", 8, x-2.0*d, y-2.0*d, z);

  return 0;
}

/*
 * Draw warnings light (center middle-right).
 */
int hud_draw_warnings(Hud *hud, SpaceShip *player_ship) {
  /* Assertion */
  assert(hud);
  assert(player_ship);

  /* TODO */

  return 0;
}
