/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "mat.h"
#include "vec3d.h"

/*
 * Make identity matrix.
 */
void ident_mat2(float* m) {
  v2_set(&m[0], 1.0, 0.0);
  v2_set(&m[1], 0.0, 1.0);
}
void ident_mat3(float* m) {
  v3_set(&m[0], 1.0, 0.0, 0.0);
  v3_set(&m[1], 0.0, 1.0, 0.0);
  v3_set(&m[1], 0.0, 0.0, 1.0);
}
void ident_mat4(float* m) {
  v4_set(&m[0], 1.0, 0.0, 0.0, 0.0);
  v4_set(&m[4], 0.0, 1.0, 0.0, 0.0);
  v4_set(&m[8], 0.0, 0.0, 1.0, 0.0);
  v4_set(&m[12], 0.0, 0.0, 0.0, 1.0);
}

/*
 * Make matrix from position/orientation.
 * Be carefull ! We use a left-handed coordinate system, like POV-Ray
 * to position right, forward and up vectors.
 */
void make_mat4(float *m, const float *xyz, 
	       const float *right, const float *forward, const float *up) {
  v4_copy(&m[0], right);
  v4_copy(&m[4], up);
  v4_scale2(&m[8], forward, -1.0);
  v4_copy(&m[12], xyz);
  m[3] = m[7] = m[11] = 0.0;
  m[15] = 1.0;
}

/*
 * Copy src in dst.
 */
void copy_mat4(float *dst, const float *src) {
  copy_vec4(&dst[0], &src[0]);
  copy_vec4(&dst[4], &src[4]);
  copy_vec4(&dst[8], &src[8]);
  copy_vec4(&dst[12], &src[12]);
}

/*
 * Scale : dst = k * src .
 */
void scale_mat4(float *dst, const float *src, const float k) {
  scale2_vec4(&dst[0], &src[0], k);
  scale2_vec4(&dst[4], &src[4], k);
  scale2_vec4(&dst[8], &src[8], k);
  scale2_vec4(&dst[12], &src[12], k);
}

/*
 * Addition : m = a + b .
 */
void add_mat4(float *m, const float *a, const float *b) {
  unsigned int i, j;
  
  for(i=0; i<4; i++) {
    for(j=0; j<4; j++) {
      m[i+j*4] = a[i+j*4] + b[i+j*4];
    }
  }
}

/*
 * Determinant.
 */
static float det_mat3(const float *m) {
  float det;

  det = m[0] * (m[4]*m[8] - m[7]*m[5])
    - m[1] * (m[3]*m[8] - m[6]*m[5])
    + m[2] * (m[3]*m[7] - m[6]*m[4]);

  return det;
}
static void sub_mat4(const float *mr, float *mb, const int i, const int j) {
  unsigned int ti, tj, idst=0, jdst=0;

  for(ti=0; ti<4; ti++) {
    if(ti<i) {
      idst = ti;
    }
    else {
      if(ti>i) {
	idst = ti-1;
      }
    }
    for(tj=0; tj<4; tj++) {
      if(tj<j) {
	jdst = tj;
      }
      else {
	if(tj>j) {
	  jdst = tj-1;
	}
      }
      if(ti!=i && tj!=j) {
	mb[idst*3+jdst] = mr[ti*4+tj];
      }
    }
  }
}
static float det_mat4(const float *m) {
  float det, result=0, i=1;
  mat3_t msub3;
  unsigned int n;

  for(n=0; n<4; n++, i*=-1) {
    sub_mat4(m, msub3, 0, n);
    det = det_mat3(msub3);
    result += m[n] * det * i;
  }
  return result;
}

/*
 * Invert.
 */
void invert_mat3(float *mr, const float *ma) {
  float det = det_mat3(ma), d;

  if(fabs(det)<=FLT_EPSILON) {
    ident_mat3(mr); /* do something */
    return;
  }
  d = 1.0 / det;

  mr[0] = (ma[4]*ma[8] - ma[5]*ma[7]) * d;
  mr[1] = - (ma[1]*ma[8] - ma[7]*ma[2]) * d;
  mr[2] = (ma[1]*ma[5] - ma[4]*ma[2]) * d;

  mr[3] = - (ma[3]*ma[8] - ma[5]*ma[6]) * d;
  mr[4] = (ma[0]*ma[8] - ma[6]*ma[2]) * d;
  mr[5] = - (ma[0]*ma[5] - ma[3]*ma[2]) * d;

  mr[6] = (ma[3]*ma[7] - ma[6]*ma[4]) * d;
  mr[7] = - (ma[0]*ma[7] - ma[6]*ma[1]) * d;
  mr[8] = (ma[0]*ma[4] - ma[1]*ma[3]) * d;

}
int invert_mat4(float *mr, const float *ma) {
  float mdet = det_mat4(ma), d;
  mat3_t mtemp;
  unsigned int i, j;
  int sign;

  if(fabs(mdet)<=FLT_EPSILON) {
    return 0;
  }
  d = 1.0 / mdet;
  
  for(i=0; i<4; i++) {
    for(j=0; j<4; j++) {
      sign = 1 - ( (i+j)%2 ) * 2;
      sub_mat4(ma, mtemp, i, j);
      mr[i+j*4] = (det_mat3(mtemp) * sign) * d;
    }
  }
  return 1;
}

/*
 * Transpose. Poor man's invert, 
 * can be used when matrix is a simple rotate-translate.
 */
void transpose_mat4(float *m) {
  mat4_t tmp;
  unsigned int i, j;
  
  copy_mat4(tmp, m);

  for(i=0; i<4; i++) {
    for(j=0; j<4; j++) {
      m[i+j*4] = tmp[j+i*4];
    }
  }
}

/*
 * Multiplication : m = a * b .
 */
void mult_mat4(float *m, const float *a, const float *b) {
  unsigned int j;

  for(j=0; j<4; j++) {
    m[j*4] = b[0] * a[j*4] 
      + b[4] * a[1+j*4]
      + b[8] * a[2+j*4]
      + b[12] * a[3+j*4];
     
    m[1+j*4] = b[1] * a[j*4] 
      + b[5] * a[1+j*4]
      + b[9] * a[2+j*4]
      + b[12] * a[3+j*4];
     
    m[2+j*4] = b[2] * a[j*4]
      + b[6] * a[1+j*4]
      + b[10] * a[2+j*4]
      + b[14] * a[3+j*4];
     
    m[3+j*4] = b[3] * a[j*4]
      + b[7] * a[1+j*4]
      + b[11] * a[2+j*4]
      + b[15] * a[3+j*4];
  }
}

/*
 * Make a translation matrix from direction.
 */
void make_trans_mat4(float *m, const float *dir) {
  m[1] = m[4] = m[8]= 0.0;
  m[2] = m[6] = m[9]= 0.0;
  m[3] = m[7] = m[11]= 0.0;
  m[0] = m[5] = m[10] = m[15] = 1.0;
  copy_vec3(&m[12], dir);
}

/*
 * Make matrix from Euler angles (in degrees).
 */
void euler_to_mat4(float *m, const float *hpr) {
  float ch, sh, cp, sp, cr, sr;
  float cpsh, spch;

  ch = m_cos(hpr[0]);
  sh = m_sin(hpr[0]);
  cp = m_cos(hpr[1]);
  sp = m_sin(hpr[1]);
  cr = m_cos(hpr[2]);
  sr = m_sin(hpr[2]);
  cpsh = cp * sh;
  spch = sp * sh;

  m[0] = ch * cr;
  m[1] = -ch * sr;
  m[2] = -sh;
  m[4] = -spch * cr + cp * sr;
  m[5] = spch * sr + cp * cr;
  m[6] = -sp * ch;
  m[8] = cpsh * cr + sp * sr;
  m[9] = -cpsh * sr + sp * cr;
  m[10] = cp * ch;

  m[3] = m[7] = m[11] = m[12] = m[13] = m[14] = 0.0;
  m[15] = 1.0;
}

/*
 * Matrix to Euler.
 */
void mat4_to_euler(float *hpr, const float *m) {
  float p, h, r; 
  float ch;
  float trx, try;

  h = -m_asin(m[2]);
  ch = m_cos(h);

  if(fabs(ch)>FLT_EPSILON) {
    float ich = 1.0 / ch;
    trx = m[10] * ich;
    try = -m[6] * ich;
    p = m_atan2(try, trx);
    trx = m[0] * ich;
    try = m[1] * ich;
    r = m_atan2(try, trx);
  }
  else {
    p = 0.0;
    trx = m[5];
    try = m[4];
    r = m_atan2(try, trx);
  }

  hpr[0] = m_clamp(h, 0, 360);
  hpr[1] = m_clamp(p, 0, 360);
  hpr[2] = m_clamp(r, 0, 360);
}

/*
 * Make matrix from position and Euler angles (in degrees).
 */
void make_coord_mat4(float *m, const float *xyz, const float *hpr) {
  euler_to_mat4(m, hpr);
  copy_vec3(&m[12], xyz);
}

/*
 * Make a rotation matrix from angle/axis : use quaternion !
 */
void make_rot_mat4(float *m, const float angle, const float *axis) {
  vec3_t ax;
  float s = m_sin(angle);
  float c = m_cos(angle);
  float t = 1.0 - c;

  v3_copy(ax, axis);
  v3_norm(ax); 
   
  m[0] = t * ax[0] * ax[0] + c;
  m[4] = t * ax[0] * ax[1] + s * ax[2];
  m[8] = t * ax[0] * ax[2] - s * ax[1];
  m[12] = 0.0;

  m[1] = t * ax[1] * ax[0] - s * ax[2];
  m[5] = t * ax[1] * ax[1] + c;
  m[9] = t * ax[1] * ax[2] + s * ax[0];
  m[13] = 0.0;

  m[2] = t * ax[2] * ax[0] + s * ax[1];
  m[6] = t * ax[2] * ax[1] - s * ax[0];
  m[10] = t * ax[2] * ax[2] + c;
  m[14] = 0.0;

  m[3] = m[7] = m[11] = 0.0;
  m[15] = 1.0;
}

/*
 * Transform vector by matrix.
 */
void xform_vec3(float *dst, const float *src, const float *m) {
  float t0 = src[0];
  float t1 = src[1];
  float t2 = src[2];

  dst[0] = t0 * m[0] + t1 * m[4] + t2 * m[8];
  dst[1] = t0 * m[1] + t1 * m[5] + t2 * m[9];
  dst[2] = t0 * m[2] + t1 * m[6] + t2 * m[10];
}
void xform_vec4(float *dst, const float *src, const float *m) {
  fprintf(stderr, "Not implemented yet!\n");
}
void fullxform_vec3(float *dst, const float *src, const float *m) {
  fprintf(stderr, "Not implemented yet!\n");
}
void fullxform_vec4(float *dst, const float *src, const float *m) {
  fprintf(stderr, "Not implemented yet!\n");
}

/*
 * Transform point by matrix.
 */
void xform_pnt3(float *dst, const float *src, const float *m) {
  float t0 = src[0];
  float t1 = src[1];
  float t2 = src[2];

  dst[0] = t0 * m[0] + t1 * m[1] + t2 * m[2] + m[3];
  dst[1] = t0 * m[4] + t1 * m[5] + t2 * m[6] + m[7];
  dst[2] = t0 * m[8] + t1 * m[9] + t2 * m[10] + m[11];
}
void xform_pnt4(float *dst, const float *src, const float *m) {
  fprintf(stderr, "Not implemented yet!\n");
}
void fullxform_pnt3(float *dst, const float *src, const float *m) {
  fprintf(stderr, "Not implemented yet!\n");
}
void fullxform_pnt4(float *dst, const float *src, const float *m) {
  fprintf(stderr, "Not implemented yet!\n");
}

/*
 * Test the libruary.
 */
void test_matrices(void) {
  vec3_t x3,u3,v3,w3;
  vec3_t xyz3,right3,forward3,up3,hpr3;
  mat4_t m4,n4,t4;

  printf("-> Testing matrix routines:\n");

  m4_ident(m4);
  printf("m4_ident(m4)= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("              [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("              [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("              [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);

  v4_set(&m4[0], 0.0, 1.0, 2.0, 3.0);
  v4_set(&m4[4], 4.0, 5.0, 6.0, 7.0);
  v4_set(&m4[8], 8.0, 9.0, 10.0, 11.0);
  v4_set(&m4[12], 12.0, 13.0, 14.0, 15.0);
  printf("m4_set(&m4[0],..)= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);

  v3_set(xyz3, 0.0, 0.0, -200.0);
  v3_set(right3, 1.0, 0.0, 0.0);
  v3_set(forward3, 0.0, 0.0, -1.0);
  v3_set(up3, 0.0, 1.0, 0.0);
  m4_make(m4, xyz3, right3, forward3, up3);
  printf("m4_make(m4,xyz3,right3,forward3,up3)= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("                                      [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("                                      [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("                                      [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);

  v3_copy(u3, right3);
  v3_copy(v3, up3);
  v3_scale2(w3, forward3, -1.0);
  printf("u3= <%.2f,%.2f,%.2f>, v3= <%.2f,%.2f,%.2f>, w3= <%.2f,%.2f,%.2f>\n", 
	 u3[0], u3[1], u3[2], v3[0], v3[1], v3[2], w3[0], w3[1], w3[2]);

  m4_transmat(n4, xyz3);
  printf("m4_transmat(n4,xyz3)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                      [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                      [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                      [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);

  m4_rotmat(n4, 60.0, u3);
  printf("m4_rotmat(n4,60.0,u3)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);
  m4_rotmat(n4, 60.0, v3);
  printf("m4_rotmat(n4,60.0,v3)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);
  m4_rotmat(n4, 60.0, w3);
  printf("m4_rotmat(n4,60.0,w3)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);

  m4_invert(t4,n4);
  printf("m4_invert(t4,n4)= [%.2f %.2f %.2f %.2f]\n", t4[0], t4[4], t4[8], t4[12]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[1], t4[5], t4[9], t4[13]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[2], t4[6], t4[10], t4[14]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[3], t4[7], t4[11], t4[15]);

  m4_copy(t4,n4);
  printf("m4_copy(t4,n4)= [%.2f %.2f %.2f %.2f]\n", t4[0], t4[4], t4[8], t4[12]);
  printf("                [%.2f %.2f %.2f %.2f]\n", t4[1], t4[5], t4[9], t4[13]);
  printf("                [%.2f %.2f %.2f %.2f]\n", t4[2], t4[6], t4[10], t4[14]);
  printf("                [%.2f %.2f %.2f %.2f]\n", t4[3], t4[7], t4[11], t4[15]);
  m4_transpose(t4);
  printf("m4_transpose(t4)= [%.2f %.2f %.2f %.2f]\n", t4[0], t4[4], t4[8], t4[12]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[1], t4[5], t4[9], t4[13]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[2], t4[6], t4[10], t4[14]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[3], t4[7], t4[11], t4[15]);

  m4_add(t4, m4, n4);
  printf("m4_add(t4,m4,n4)= [%.2f %.2f %.2f %.2f]\n", t4[0], t4[4], t4[8], t4[12]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[1], t4[5], t4[9], t4[13]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[2], t4[6], t4[10], t4[14]);
  printf("                  [%.2f %.2f %.2f %.2f]\n", t4[3], t4[7], t4[11], t4[15]);
   
  m4_mult(t4, m4, n4);
  printf("m4_mult(t4,m4,n4)= [%.2f %.2f %.2f %.2f]\n", t4[0], t4[4], t4[8], t4[12]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", t4[1], t4[5], t4[9], t4[13]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", t4[2], t4[6], t4[10], t4[14]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", t4[3], t4[7], t4[11], t4[15]);

  v3_set(hpr3, 0.0, 0.0, 60.0);
  printf("v3_set(hpr3,0.0,0.0,60.0)= <%.2f,%.2f,%.2f>\n", hpr3[0], hpr3[1], hpr3[2]);
  m4_fromeuler(t4, hpr3);
  printf("m4_fromeuler(t4,hpr3)= [%.2f %.2f %.2f %.2f]\n", t4[0], t4[4], t4[8], t4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", t4[1], t4[5], t4[9], t4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", t4[2], t4[6], t4[10], t4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", t4[3], t4[7], t4[11], t4[15]);
  m4_euler(hpr3, t4);
  printf("m4_euler(hpr3,t4)= <%.2f,%.2f,%.2f>\n", hpr3[0], hpr3[1], hpr3[2]);

  m4_xformvec3(x3, u3, n4);
  printf("m4_xformvec3(x3, u3, n4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  m4_xformvec3(x3, v3, n4);
  printf("m4_xformvec3(x3, v3, n4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  m4_xformvec3(x3, w3, n4);
  printf("m4_xformvec3(x3, w3, n4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);

  m4_xformvec3(x3, u3, t4);
  printf("m4_xformvec3(x3, u3, t4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  m4_xformvec3(x3, v3, t4);
  printf("m4_xformvec3(x3, v3, t4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  m4_xformvec3(x3, w3, t4);
  printf("m4_xformvec3(x3, w3, t4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);

}
