/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* 
 * Simple matrix routines. This file is meant to be included 
 * in a .c file, hence all the routines are defined statically.
 * A good C compiler should be able to automatically inline all
 * these routines. 
 */

#ifndef _MAT_H
#define _MAT_H

/*
 * Matrix ().
 * m4 = [ 0 4 8 12 ]
 *      [ 1 5 9 13 ]
 *      [ 2 6 10 14 ]
 *      [ 3 7 11 15 ]
 * m4[li][col] = m4[li+col*4]
 */
typedef float mat2_t[4];
typedef float mat3_t[9];
typedef float mat4_t[16];

/*
 * Make identity matrix.
 */
void ident_mat2(float* m);
void ident_mat3(float* m);
void ident_mat4(float* m);
#define m2_ident ident_mat2
#define m3_ident ident_mat3
#define m4_ident ident_mat4

/*
 * Make matrix from position/orientation.
 * Be carefull ! We use a left-handed coordinate system, like POV-Ray
 * to position right, forward and up vectors.
 */
void make_mat4(float *m, const float *xyz, 
               const float *right, const float *forward, const float *up);
#define m4_make make_mat4

/*
 * Copy src in dst.
 */
void copy_mat4(float *dst, const float *src);
#define m4_copy copy_mat4

/*
 * Scale : dst = k * src .
 */
void scale_mat4(float *dst, const float *src, const float k);
#define m4_scale scale_mat4

/*
 * Addition : m = a + b .
 */
void add_mat4(float *m, const float *a, const float *b);
#define m4_add add_mat4

/*
 * Invert.
 */
void invert_mat3(float *mr, const float *ma);
int invert_mat4(float *mr, const float *ma);
#define m3_invert invert_mat3
#define m4_invert invert_mat4

/*
 * Transpose. Poor man's invert, 
 * can be used when matrix is a simple rotate-translate.
 */
void transpose_mat4(float *m);
#define m4_transpose transpose_mat4

/*
 * Multiplication : m = a * b .
 */
void mult_mat4(float *m, const float *a, const float *b);
#define m4_mult mult_mat4

/*
 * Make a translation matrix from direction.
 */
void make_trans_mat4(float *m, const float *dir);
#define m4_transmat make_trans_mat4

/*
 * Make matrix from Euler angles (in degrees).
 */
void euler_to_mat4(float *m, const float *hpr);
#define m4_fromeuler euler_to_mat4

/*
 * Matrix to Euler.
 */
void mat4_to_euler(float *hpr, const float *m);
#define m4_euler mat4_to_euler

/*
 * Make matrix from position and Euler angles (in degrees).
 */
void make_coord_mat4(float *m, const float *xyz, const float *hpr);
#define m4_makecoord make_coord_mat4

/*
 * Make a rotation matrix from angle/axis : use quaternion !
 */
void make_rot_mat4(float *m, const float angle, const float *axis);
#define m4_rotmat make_rot_mat4

/*
 * Transform vector by matrix.
 */
void xform_vec3(float *dst, const float *src, const float *m);
void xform_vec4(float *dst, const float *src, const float *m);
void fullxform_vec3(float *dst, const float *src, const float *m);
void fullxform_vec4(float *dst, const float *src, const float *m);
#define m4_xformvec3 xform_vec3
#define m4_xformvec4 xform_vec4
#define m4_fullxformvec3 fullxform_vec3
#define m4_fullxformvec4 fullxform_vec4

/*
 * Transform point by matrix.
 */
void m4_xform_pnt3(float *dst, const float *src, const float *m);
void m4_xform_pnt4(float *dst, const float *src, const float *m);
void m4_fullxform_pnt3(float *dst, const float *src, const float *m);
void m4_fullxform_pnt4(float *dst, const float *src, const float *m);
#define m4_xformpnt3 xform_pnt3
#define m4_xformpnt4 xform_pnt4
#define m4_fullxformpnt3 fullxform_pnt3
#define m4_fullxformpnt4 fullxform_pnt4

/*
 * Test the libruary.
 */
void test_matrices(void);
#define m4_test test_matrices


#endif /* _MAT_H */
