/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>

#include <assert.h>

#include "sdlib.h"

#include "modeldesc.h"

#include "vec3d.h"
#include "xml.h"

/*
 * Constructor.
 */
ModelDesc *model_desc_new(void) {
  ModelDesc *model_desc = (ModelDesc *)malloc(sizeof(ModelDesc));
  if(!model_desc) {
    sd_error("Can't allocate memory (ModelDesc)!");
  }
  
  /* Identification */
  sprintf(model_desc->name, "None");
  sprintf(model_desc->class, "Unknown");
  sprintf(model_desc->category, "Unknown");
  sprintf(model_desc->type, "Unknown");
  model_desc->camp = UNKNOWN;

  /* Description */
  sprintf(model_desc->manoeuvrability, "Unknown");
  sprintf(model_desc->power, "Unknown");
  sprintf(model_desc->range, "Unknown");
  sprintf(model_desc->autonomy, "Unknown");
  sprintf(model_desc->crew, "Unknown");
  sprintf(model_desc->tonnage, "Unknown");
  sprintf(model_desc->freight, "Unknown");

  model_desc->obj = obj_new(); /* object */
  model_desc->length = 0;
  model_desc->width = 0;
  model_desc->height = 0;
  model_desc->radius = 0;
   
  /* Caracteristics */
  model_desc->hit_points = 0;
  model_desc->hull = 0;
  model_desc->armour = 0;

  /* Max linear acceleration (m/s/s) */
  model_desc->forward_accel = 0.0;
  model_desc->back_accel = 0.0;
  model_desc->lateral_accel = 0.0; /* right/left */
  model_desc->up_accel = 0.0;
  model_desc->down_accel = 0.0;
  /* Max linear speed (m/s) */
  model_desc->forward_velocity = 0.0;
  model_desc->back_velocity = 0.0;
  model_desc->lateral_velocity = 0.0; /* right/left */
  model_desc->up_velocity = 0.0;
  model_desc->down_velocity = 0.0;
  
  /* Max angular rotation acceleration (deg/s/s) */
  model_desc->yaw_accel = 0.0; /* or heading */
  model_desc->pitch_accel = 0.0;
  model_desc->roll_accel = 0.0;
  /* Max angular rotation speed (deg/s) */
  model_desc->yaw_velocity = 0.0; /* or heading */
  model_desc->pitch_velocity = 0.0;
  model_desc->roll_velocity = 0.0;

  return model_desc;
}

/*
 * Destructor.
 */
int model_desc_destroy(ModelDesc *model_desc) {
  /* Assertion */
  assert(model_desc);
  assert(model_desc->obj);

  /* Free memory */
  obj_destroy(model_desc->obj);
  free(model_desc);

  return 0;
}

/*
 * Initialization.
 */
int model_desc_read_xml(ModelDesc *model_desc, const char *model_file) {
  Xml *model = NULL;
  char *val = NULL;

  /* Assertion */
  assert(model_desc);
  assert(model_desc->obj);

  /* Load model_desc */
  model = xml_new();
  printline("\nParsing model file: %s\n", model_file);
  xml_init(model, model_file);

  /* Parse xml file */

  /* Identification */
  val = xml_search_path(model, "/Stardust/ship/id/name/text()");
  sprintf(model_desc->name, val);
  val = xml_search_path(model, "/Stardust/ship/id/class/text()");
  sprintf(model_desc->class, val);
  val = xml_search_path(model, "/Stardust/ship/id/category/text()");
  sprintf(model_desc->category, val);
  val = xml_search_path(model, "/Stardust/ship/id/type/text()");
  sprintf(model_desc->type, val);
  val = xml_search_path(model, "/Stardust/ship/id/camp/text()");
  if(!strcmp(val, "UNKNOWN") || !strcmp(val, "UNK")) {
    model_desc->camp = UNK;
  }
  else if(!strcmp(val, "NEUTRAL") || !strcmp(val, "NEU")) {
    model_desc->camp = NEU;
  }
  else if(!strcmp(val, "NAVY") || !strcmp(val, "FEDERATION") || !strcmp(val, "NVY")) {
    model_desc->camp = NVY;
  }
  else if(!strcmp(val, "ENNEMY") || !strcmp(val, "ALLIANCE") || !strcmp(val, "ENY")) {
    model_desc->camp = ENY;
  }
  else {
    model_desc->camp = UNK;
  }

  /* Description */
  val = xml_search_path(model, "/Stardust/ship/desc/manoeuvrability/text()");
  sprintf(model_desc->manoeuvrability, val);
  val = xml_search_path(model, "/Stardust/ship/desc/power/text()");
  sprintf(model_desc->power, val);
  val = xml_search_path(model, "/Stardust/ship/desc/range/text()");
  sprintf(model_desc->range, val);
  val = xml_search_path(model, "/Stardust/ship/desc/autonomy/text()");
  sprintf(model_desc->autonomy, val);
  val = xml_search_path(model, "/Stardust/ship/desc/crew/text()");
  sprintf(model_desc->crew, val);
  /*val = xml_search_path(model, "/Stardust/ship/desc/length/text()");*/
  val = xml_search_path(model, "/Stardust/ship/desc/tonnage/text()");
  sprintf(model_desc->tonnage, val);
  val = xml_search_path(model, "/Stardust/ship/desc/freight/text()");
  sprintf(model_desc->freight, val);  

  val = xml_search_path(model, "/Stardust/ship/obj/file/text()");
  obj_init(model_desc->obj, val);
  {
    /* Get model properties */
    vec3_t len;
    sub_vec3(len, obj_max(model_desc->obj), obj_min(model_desc->obj));
    model_desc->length = len[2];
    model_desc->width = len[0];
    model_desc->height = len[1];
    model_desc->radius = obj_radius(model_desc->obj);
  }

  /* Caracteristics */
  val = xml_search_path(model, "/Stardust/ship/caract/hit_points/text()");
  model_desc->hit_points = atof(val);
  val = xml_search_path(model, "/Stardust/ship/def/hull/text()");
  model_desc->hull = atof(val);
  val = xml_search_path(model, "/Stardust/ship/def/armour/text()");
  model_desc->armour = atof(val);

  /* Max linear acceleration (m/s/s) */
  val = xml_search_path(model, "/Stardust/ship/accel/forward/text()");
  model_desc->forward_accel = atof(val);
  val = xml_search_path(model, "/Stardust/ship/accel/back/text()");
  model_desc->back_accel = atof(val);
  val = xml_search_path(model, "/Stardust/ship/accel/lat/text()");
  model_desc->lateral_accel = atof(val); /* right/left */
  val = xml_search_path(model, "/Stardust/ship/accel/up/text()");
  model_desc->up_accel = atof(val);
  val = xml_search_path(model, "/Stardust/ship/accel/down/text()");
  model_desc->down_accel = atof(val);
  /* Max linear speed (m/s) */
  val = xml_search_path(model, "/Stardust/ship/velocity/forward/text()");
  model_desc->forward_velocity = atof(val);
  val = xml_search_path(model, "/Stardust/ship/velocity/back/text()");
  model_desc->back_velocity = atof(val);
  val = xml_search_path(model, "/Stardust/ship/velocity/lat/text()");
  model_desc->lateral_velocity = atof(val); /* right/left */
  val = xml_search_path(model, "/Stardust/ship/velocity/up/text()");
  model_desc->up_velocity = atof(val);
  val = xml_search_path(model, "/Stardust/ship/velocity/down/text()");
  model_desc->down_velocity = atof(val);
  
  /* Max angular rotation acceleration (deg/s/s) */
  val = xml_search_path(model, "/Stardust/ship/accel/yaw/text()");
  model_desc->yaw_accel = atof(val); /* or heading */
  val = xml_search_path(model, "/Stardust/ship/accel/pitch/text()");
  model_desc->pitch_accel = atof(val);
  val = xml_search_path(model, "/Stardust/ship/accel/roll/text()");
  model_desc->roll_accel = atof(val);
  /* Max angular rotation speed (deg/s) */
  val = xml_search_path(model, "/Stardust/ship/velocity/yaw/text()");
  model_desc->yaw_velocity = atof(val); /* or heading */
  val = xml_search_path(model, "/Stardust/ship/velocity/pitch/text()");
  model_desc->pitch_velocity = atof(val);
  val = xml_search_path(model, "/Stardust/ship/velocity/roll/text()");
  model_desc->roll_velocity = atof(val);

  /* Free memory */
  xml_destroy(model);

  return 0;
}

/*
 * Draw.
 */
int model_desc_draw(ModelDesc *model_desc) {
  /* Assertion */
  assert(model_desc);
  assert(model_desc->obj);

  /* Draw model */
  obj_draw(model_desc->obj, DISPLAY_LIST);

  return 0;
}
