/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <malloc.h>
#include <math.h>

#include <assert.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include <zlib.h> /* 'zlib' compression library */

#include "sdlib.h"

#include "obj.h"

#include "vec3d.h"

/* Useful macros */
#define intpar(str, x) sscanf(str, "%*s %d", &(x))
#define int3par(str, x1, x2, x3) sscanf(str, "%*s %d %d %d", &(x1), &(x2), &(x3))

#define dblpar(str, x) sscanf(str, "%*s %f", &(x))
#define dbl3par(str, x, y, z) sscanf(str, "%*s %f%f%f", &(x), &(y), &(z))

#define MAX(a,b) (((a) < (b)) ? b : a)
#define MIN(a,b) (((a) > (b)) ? b : a)

/* A simple Face definition (polygon) */
struct face { 
  int ptcount; /* nb vertices for this face */ 
  int *v;  /* idxs on vertices */
  int *vt; /* idxs on texcoords */
  int *vn; /* idxs on normals */
  vec3_t n; /* normal of the face */
};

/* Color definition */
struct color { 
  float Ka[4], Kd[4], Ks[4]; /* ambiant, diffuse and specular colors  */
  float Ns; /* shininess */
};

/*
 * Constructor.
 */
Obj *obj_new(void) {
  /* Allocation de la structure */
  Obj *obj = (Obj *)malloc(sizeof(Obj));
  if(!obj) {
    sd_error("Can't allocate memory (Obj)!");
  }

  /* Obj init */
  sprintf(obj->name, "None");
  obj->nbvertices = 0;
  obj->nbfaces = 0;

  obj->id = 0;

  obj->texture = texture_new();

  obj->inter = NULL;
  obj->indices = NULL;

  obj->sizemax = 0.0;
  zero_vec3(obj->centre);
  zero_vec3(obj->barycentre);

  zero_vec3(obj->min);
  zero_vec3(obj->max);

  obj->radius = 0.0;

  return obj;
}

/*
 * Destructor.
 */
int obj_destroy(Obj *obj) {
  /* Assertion */
  assert(obj);
  assert(obj->texture);

  /* Free memory */
  texture_destroy(obj->texture);
  if(obj->inter) {
    free(obj->inter);
  }
  if(obj->indices) {
    free(obj->indices);
  }
  free(obj);

  return 0;
}

/*
 * Parse faces idxs.
 */
#define V_PAR   0
#define VT_PAR  1
#define VN_PAR  2
#define VTN_PAR 3
static short intnpar(char *line, int mode, struct face *face) {
  char *str = strtok(line, " \n");
  int pt = 0;

  /* Init tab */
  face->ptcount = 3;
  face->v = (int *)malloc((face->ptcount)*sizeof(int));
  face->vt = (int *)malloc((face->ptcount)*sizeof(int));
  face->vn = (int *)malloc((face->ptcount)*sizeof(int));

  /* Pass "f" */
  str = strtok(NULL, " \n");

  while(str) {
    if(pt >= face->ptcount) { /* realloc if necessary */
      face->ptcount = pt+1;
      face->v = (int *)realloc(face->v, (face->ptcount)*sizeof(int));
      face->vt = (int *)realloc(face->vt, (face->ptcount)*sizeof(int));
      face->vn = (int *)realloc(face->vn, (face->ptcount)*sizeof(int));
    }
    
    if(mode == V_PAR) { /* f v1 v2 v3 ... */
      sscanf(str, "%d", &(face->v[pt]));
      face->vn[pt] = face->v[pt];
    }
    else if(mode == VT_PAR) { /* f v1/vt1 v2/vt2 v3/vt3 ... */
      sscanf(str, "%d/%d", &(face->v[pt]), &(face->vt[pt]));
      face->vn[pt] = face->v[pt];
    }
    else if(mode == VN_PAR) { /* f v1//vn1 v2//vn2 v3//vn3 ... */
      sscanf(str, "%d//%d", &(face->v[pt]), &(face->vn[pt]));
    }
    else if(mode == VTN_PAR) { /* f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 ... */
      sscanf(str, "%d/%d/%d", 
	     &(face->v[pt]), &(face->vt[pt]), &(face->vn[pt]));
    }
    else {
      sd_warning("Must never happen!");
    }

    /* Next token */
    str = strtok(NULL, " \n");
    pt++;
  }

  return 1;
} 

/* 
 * Load a Wavefront OBJ file.
 * Return the display list idx.
 */
#define ALLOC_MORE 100
GLuint obj_load(Obj *object, const char *obj_file) {
  gzFile *gzp = NULL;
  char line[1024], command[1024];

  char filename_mtl[256] = "\0";
  GLuint obj_texid = 0;

  /* Vertices */
  int vcount = 0, vmax = ALLOC_MORE;
  vec3_t *tabvectors = NULL;
  /* Normals */
  int vncount = 0, vnmax = ALLOC_MORE;
  vec3_t *tabnormals = NULL;
  /* Colors */
  int /*ccount = 0,*/ cmax = ALLOC_MORE;
  struct color *tabcolors = NULL;
  /* TexCoords */
  int vtcount = 0, vtmax = ALLOC_MORE;
  vec3_t *tabtexcoords = NULL;
  /* Faces */
  int fcount = 0, fmax = ALLOC_MORE;
  struct face *tabfaces = NULL;

  vec3_t u, v;
  int firstindex = 1;
  int i, j;
  vec3_t g, min, max;

  /* Assertion */
  assert(object);

  /* Alloc the room for all vectors, normals, colors and tex coords */
  tabvectors = (vec3_t *)malloc((vmax+1)*sizeof(vec3_t));
  assert(tabvectors);
  tabnormals = (vec3_t *)malloc((vnmax+1)*sizeof(vec3_t));
  assert(tabnormals);
  tabcolors = (struct color *)malloc((cmax+1)*sizeof(struct color));
  assert(tabcolors);
  tabtexcoords = (vec3_t *)malloc((vtmax+1)*sizeof(vec3_t));
  assert(tabtexcoords);

  /* Alloc the room for all the triangular faces */
  tabfaces = (struct face *)malloc((fmax+1)*sizeof(struct face));
  assert(tabfaces);

  printline("\nLoading Wavefront OBJ: %s\n", obj_file);
  if(!(gzp = gzopen(obj_file,"rb"))) {
    char obj_file2[FILENAME_LGTH];

    dprintf("Trying in directory: "DATA_DIR"\n");
    strcpy(obj_file2, DATA_DIR);
    strcat(obj_file2, obj_file);

    if(!(gzp = gzopen(obj_file2,"rb"))) {
      sd_error("Can't open file(OBJ): %s", obj_file);
    }
  }
  while(gzgets(gzp, line, 1024) != Z_NULL) {

    if((line[0] != '#') && (line[0] != '\n')) {
      if(sscanf(line, "%s", command)) {
	
        if(!strcmp(command, "o")) { /* object name */
	  char obj_name[32];
	  sscanf(line, "%*s %s", obj_name);           
        }
        else if(!strcmp(command, "mtllib")) { /* material lib filename.mtl */
	  sscanf(line, "%*s %s", filename_mtl);           
	  sd_warning("Not yet supported (%s)!", filename_mtl);
        }
        else if(!strcmp(command, "v")) { /* vertices */
	  if(vcount >= vmax) { /* realloc if necessary */
	    vmax += ALLOC_MORE;
	    tabvectors = (vec3_t *)realloc(tabvectors, 
					   (vmax+1)*sizeof(vec3_t));
	    assert(tabvectors);
	  }
          /* Read the vertex */
          dbl3par(line,
                  tabvectors[vcount][0],
                  tabvectors[vcount][1],
                  tabvectors[vcount][2]);
          vcount++;
	}
        else if(!strcmp(command, "vt")) { /* texcoords */
	  if(vtcount >= vtmax) { /* realloc if necessary */
	    vtmax += ALLOC_MORE;
	    tabtexcoords = (vec3_t *)realloc(tabtexcoords, 
					     (vtmax+1)*sizeof(vec3_t));
	    assert(tabtexcoords);
	  }
          /* Read the texcoord */
          dbl3par(line,
                  tabtexcoords[vtcount][0],
                  tabtexcoords[vtcount][1],
                  tabtexcoords[vtcount][2]);
          vtcount++;
        }
        else if(!strcmp(command, "vn")) { /* normals */
	  if(vncount >= vnmax) { /* realloc if necessary */
	    vnmax += ALLOC_MORE;
	    tabnormals = (vec3_t *)realloc(tabnormals, 
					   (vnmax+1)*sizeof(vec3_t));
	    assert(tabnormals);
	  }
          /* Read the normal */
          dbl3par(line,
                  tabnormals[vncount][0],
                  tabnormals[vncount][1],
                  tabnormals[vncount][2]);
          vncount++;
        }
        else if(!strcmp(command, "g")) { /* object name */
	  sscanf(line, "%*s %s", object->name);           
	  printline("Loading %s group ...\n", object->name);
        }
        else if(!strcmp(command, "usemtl")) { /* material */
	  char mtl[32];
	  sscanf(line, "%*s %s", mtl);
	  /*setMaterial(filename_mtl, mtl);*/
	  sd_warning("Not yet supported (%s)!", mtl);
        }
        else if(!strcmp(command, "tex")) { /* texture */
	  char tex_file[256];
	  char img_file[256];
	  strcpy(img_file, "data/img/");
	  sscanf(line, "%*s %s", tex_file);
	  strcat(img_file, tex_file);
	  /* Loading texture */
	  obj_texid = texture_init(object->texture, img_file);
	}
        else if(!strcmp(command, "f")) { /* faces */
	  if(fcount >= fmax) { /* realloc if necessary */
	    fmax += ALLOC_MORE;
	    tabfaces = (struct face *)realloc(tabfaces, 
					      (fmax+1)*sizeof(struct face));
	    assert(tabfaces);
	  }
	  
	  /* Assume that v, vn or vt coords have been already computed
	     before reading faces v, v/vt, v//vn or v/vt/vn ! */
			       
	  /* Read the triangular face as vertex indexes */
	  if(vncount == 0 && vtcount == 0) { 
            /* f v1 v2 v3 ... */
	    intnpar(line, V_PAR, &tabfaces[fcount]);
	  }
	  else if(vncount == 0 && vtcount !=0) {
            /* f v1/vt1 v2/vt2 v3/vt3 ... */
	    intnpar(line, VT_PAR, &tabfaces[fcount]);
	  }
	  else if(vncount != 0 && vtcount == 0) {
            /* f v1//vn1 v2//vn2 v3//vn3 ... */
	    intnpar(line, VN_PAR, &tabfaces[fcount]);
	  }
          else if(vncount != 0 && vtcount != 0) {
            /* f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 ... */
	    intnpar(line, VTN_PAR, &tabfaces[fcount]);
          }
	  else {
	    sd_warning("Must never happen!");
	  }

	  /* Compute normals ? */
          if(vncount == 0) {
	    if(fcount == 0) {
	      /* Assume vertex already computed before reading faces */
	      printline("Compute normals ...\n");
	      vnmax = vcount;
	      tabnormals = (vec3_t *)realloc(tabnormals, 
					     (vnmax+1)*sizeof(vec3_t));
	      assert(tabnormals);
	      for(i=0; i<(vnmax+1); i++) {
		zero_vec3(tabnormals[i]);
	      }
	    }
            /* Calcul normal of the surface (first 3 pts) */
	    sub_vec3(u, tabvectors[tabfaces[fcount].v[1]], 
		     tabvectors[tabfaces[fcount].v[0]]);
	    sub_vec3(v, tabvectors[tabfaces[fcount].v[2]], 
		     tabvectors[tabfaces[fcount].v[0]]);
	    cross_vec3(tabfaces[fcount].n, v, u);
	    norm_vec3(tabfaces[fcount].n);
	    for(j=0; j<tabfaces[fcount].ptcount; j++) {
	      add_vec3(tabnormals[tabfaces[fcount].vn[j]],
		       tabnormals[tabfaces[fcount].vn[j]],
		       tabfaces[fcount].n);
	    }
          } /* end if vncount = 0 */

	  fcount++;
	}
        else if(!strcmp(command, "first_index")) { /* indexs begin at 0, 1 ? */
          intpar(line, firstindex);
        }
	else {
	  sd_warning("Command %s not recognized!", command);
	}

      }
    }

  } /* end while */

  gzclose(gzp); /* close file */

  /* Normalize normals computed before */
  if(vncount == 0) { 
    printline("Normalize normals ...\n");
    vncount = vcount; 
    for(i=0; i<vncount; i++)
      norm_vec3(tabnormals[i]);
  }
  
  /* Update object properties */
  object->nbvertices = vcount;
  object->nbfaces = fcount;

  /* Hack for file where indexs begin at 1 and not 0 */
  if(firstindex) {
    for(i=0; i<fcount; i++) {
      for(j=0; j<tabfaces[i].ptcount; j++) {
	tabfaces[i].v[j]--;
	tabfaces[i].vt[j]--;
	tabfaces[i].vn[j]--;
      }
    }
#if 0    
    for(i=1; i<(vncount+1); i++) {
      tabnormals[i-1][0] = tabnormals[i][0];
      tabnormals[i-1][1] = tabnormals[i][1];
      tabnormals[i-1][2] = tabnormals[i][2];
    }
#endif
  }

  /* Compute oject properties, bbox & bsphere */
  /* Barycentre */
  for(i=0, g[0]=g[1]=g[2]=0; i<vcount; i++) {
    g[0] += tabvectors[i][0];
    g[1] += tabvectors[i][1];
    g[2] += tabvectors[i][2];
  }
  object->barycentre[0] = g[0] / vcount;
  object->barycentre[1] = g[1] / vcount;
  object->barycentre[2] = g[2] / vcount;
  /* Min/Max size */
  min[0]=tabvectors[0][0];
  min[1]=tabvectors[0][0];
  min[2]=tabvectors[0][0];
  max[0]=tabvectors[0][0];
  max[1]=tabvectors[0][0];
  max[2]=tabvectors[0][0];
  for(i=1; i<vcount; i++) {
    min[0]=MIN(min[0], tabvectors[i][0]);
    min[1]=MIN(min[1], tabvectors[i][1]);
    min[2]=MIN(min[2], tabvectors[i][2]);
    max[0]=MAX(max[0], tabvectors[i][0]);
    max[1]=MAX(max[1], tabvectors[i][1]);
    max[2]=MAX(max[2], tabvectors[i][2]);
  }
  object->min[0] = min[0]; 
  object->min[1] = min[1]; 
  object->min[2] = min[2]; 
  object->max[0] = max[0]; 
  object->max[1] = max[1]; 
  object->max[2] = max[2]; 
  object->sizemax = MAX(MAX(max[0]-min[0], max[1]-min[1]), max[2]-min[2]);
  /* Center */
  object->centre[0] = (object->max[0] + object->min[0]) / 2.0; 
  object->centre[1] = (object->max[1] + object->min[1]) / 2.0; 
  object->centre[2] = (object->max[2] + object->min[2]) / 2.0; 
  /* Radius */
  object->radius = object->sizemax / 2.0;

  /* Display List */
  object->id = glGenLists(1);
  glNewList(object->id, GL_COMPILE);

  /*glDisable(GL_LIGHTING);*/
  glEnable(GL_TEXTURE_2D);

  glColor4f(0.7, 0.7, 0.7, 1.0);
  glBindTexture(GL_TEXTURE_2D, obj_texid);

  for(i=0; i<fcount; i++) {
    glBegin(GL_TRIANGLE_FAN);
    /*glNormal3fv(tabfaces[i].n);*/
    for(j=0; j<tabfaces[i].ptcount; j++) {
      if(vtcount) { glTexCoord2fv(tabtexcoords[tabfaces[i].vt[j]]); }
      glNormal3fv(tabnormals[tabfaces[i].vn[j]]);
      glVertex3fv(tabvectors[tabfaces[i].v[j]]);
    }
    glEnd();
  }

  glDisable(GL_TEXTURE_2D);
  /*glEnable(GL_LIGHTING);*/

  glEndList();

#if 0
  /* Vertex Arrays */
  object->inter = (GLfloat *)malloc((3*8*fcount)*sizeof(GLfloat));
  assert(object->inter);
  object->indices = (GLuint *)malloc(fcount*sizeof(GLuint));
  assert(object->indices);
  for(i=0, j=0; j<fcount; j++) {
    object->indices[j] = j;
    /* Point 1 */
    object->inter[i]   = 0.0;
    object->inter[i+1] = 0.0;
    object->inter[i+2] = tabnormals[tabfaces[j].v1][0];
    object->inter[i+3] = tabnormals[tabfaces[j].v1][1];
    object->inter[i+4] = tabnormals[tabfaces[j].v1][2];
    object->inter[i+5] = tabvectors[tabfaces[j].v1][0];
    object->inter[i+6] = tabvectors[tabfaces[j].v1][1];
    object->inter[i+7] = tabvectors[tabfaces[j].v1][2];
    i+=8;
    /* Point 2 */
    object->inter[i]   = 0.0;
    object->inter[i+1] = 0.0;
    object->inter[i+2] = tabnormals[tabfaces[j].v2][0];
    object->inter[i+3] = tabnormals[tabfaces[j].v2][1];
    object->inter[i+4] = tabnormals[tabfaces[j].v2][2];
    object->inter[i+5] = tabvectors[tabfaces[j].v2][0];
    object->inter[i+6] = tabvectors[tabfaces[j].v2][1];
    object->inter[i+7] = tabvectors[tabfaces[j].v2][2];
    i+=8;
    /* Point 3 */
    object->inter[i]   = 0.0;
    object->inter[i+1] = 0.0;
    object->inter[i+2] = tabnormals[tabfaces[j].v3][0];
    object->inter[i+3] = tabnormals[tabfaces[j].v3][1];
    object->inter[i+4] = tabnormals[tabfaces[j].v3][2];
    object->inter[i+5] = tabvectors[tabfaces[j].v3][0];
    object->inter[i+6] = tabvectors[tabfaces[j].v3][1];
    object->inter[i+7] = tabvectors[tabfaces[j].v3][2];
    i+=8;
  }
  glInterleavedArrays(GL_T2F_N3F_V3F, 0, &object->inter);
#endif

  /* Release memory */
  free(tabvectors);
  free(tabnormals);
  free(tabcolors);
  free(tabtexcoords);
  for(i=0; i<fcount; i++) {
    free(tabfaces[i].v);
    free(tabfaces[i].vt);
    free(tabfaces[i].vn);
  }
  free(tabfaces);

  printline(" %d vertices %d faces\n", 
	  object->nbvertices, object->nbfaces);
  printline(" length %.1f width %.1f height %.1f\n", 
	  max[2]-min[2], max[0]-min[0],  max[1]-min[1]);
  return object->id;
}

/*
 * Draw a Wavefront OBJ.
 * With method= DisplayList | VertexArray.
 */
int obj_draw(Obj *obj, int method) {
  /* Assertion */
  assert(obj);

  glDisable(GL_CULL_FACE); /* test */

  if(method == DISPLAY_LIST) { /* display list */
    if(glIsList(obj->id)) {
      glPushMatrix();
      glCallList(obj->id);
      glPopMatrix();
    }
    else {
      sd_warning("Display List not available!");
    }
  }
  else if(method == VERTEX_ARRAY) { /* vertex array */
    glPushMatrix();
    glDrawElements(GL_TRIANGLES, obj->nbfaces,
		   GL_UNSIGNED_INT, obj->indices);
    glPopMatrix();
  }

  return 0;
}

/*
 * Draw a parallelepiped (center @<0,0,0>).
 */
GLuint obj_draw_parallelepiped(float length, float width, float height, 
			       short color) {
  float l = length / 2.0;
  float w = width / 2.0;
  float h = height / 2.0;
  int id = 0;
  GLfloat col[4];

  /* Determine color (camp) */
  switch(color) {
  case YELLOW:
    set_vec4(col, 1.0, 1.0, 0.0, 1.0);
    break;
  case GREEN:
    set_vec4(col, 0.0, 1.0, 0.0, 1.0);
    break;
  case BLUE:
    set_vec4(col, 0.0, 0.0, 1.0, 1.0);
    break;
  case RED:
    set_vec4(col, 1.0, 0.0, 0.0, 1.0);
    break;
  default:
    break;
  }

  /* Generate display list */
  id = glGenLists(1);
  glNewList(id, GL_COMPILE);

  glColor4fv(col);

  glBegin(GL_QUADS);

  glNormal3f(0.0, 0.0, -1.0); /* back */
  glTexCoord2f(0.0, 0.0); glVertex3f(-w, -h, -l);
  glTexCoord2f(0.0, 1.0); glVertex3f(-w, h, -l);
  glTexCoord2f(1.0, 1.0); glVertex3f(w, h, -l);
  glTexCoord2f(1.0, 0.0); glVertex3f(w, -h, -l);

  glNormal3f(0.0, 0.0, 1.0); /* front */
  glTexCoord2f(0.0, 0.0); glVertex3f(-w, -h, l);
  glTexCoord2f(1.0, 0.0); glVertex3f(w, -h, l);
  glTexCoord2f(1.0, 1.0); glVertex3f(w, h, l);
  glTexCoord2f(0.0, 1.0); glVertex3f(-w, h, l);

  glNormal3f(-1.0, 0.0, 0.0); /* left */
  glTexCoord2f(0.0, 0.0); glVertex3f(-w, -h, -l);
  glTexCoord2f(0.0, 1.0); glVertex3f(-w, -h, l);
  glTexCoord2f(1.0, 1.0); glVertex3f(-w, h, l);
  glTexCoord2f(1.0, 0.0); glVertex3f(-w, h, -l);

  glNormal3f(1.0, 0.0, 0.0); /* right */
  glTexCoord2f(0.0, 0.0); glVertex3f(w, -h, -l);
  glTexCoord2f(1.0, 0.0); glVertex3f(w, h, -l);
  glTexCoord2f(1.0, 1.0); glVertex3f(w, h, l);
  glTexCoord2f(0.0, 1.0); glVertex3f(w, -h, l);

  glNormal3f(0.0, 1.0, 0.0); /* up */
  glTexCoord2f(0.0, 0.0); glVertex3f(-w, h, -l);
  glTexCoord2f(0.0, 1.0); glVertex3f(-w, h, l);
  glTexCoord2f(1.0, 1.0); glVertex3f(w, h, l);
  glTexCoord2f(1.0, 0.0); glVertex3f(w, h, -l);

  glNormal3f(0.0, -1.0, 0.0); /* down */
  glTexCoord2f(0.0, 0.0); glVertex3f(-w, -h, -l);
  glTexCoord2f(1.0, 0.0); glVertex3f(w, -h, -l);
  glTexCoord2f(1.0, 1.0); glVertex3f(w, -h, l);
  glTexCoord2f(0.0, 1.0); glVertex3f(-w, -h, l);

  glEnd();
  
  glEndList();

  return id;
};
