/*
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "quat.h"
#include "vec3d.h"
#include "mat.h"

/*
 * Make identity quaternion.
 */
void ident_quat(float *q) {
  zero_vec3(q);
  q[3] = 1.0;
}

/*
 * Set quaternion.
 */
void init_quat(float *q, const float x, const float y, const float z, const float w) {
  set_vec4(q, x, y, z, w);
}

/*
 * Copy src in dst.
 */
void copy_quat(float *dst, const float *src) {
  copy_vec4(dst, src);
}

/*
 * Invert.
 */
void invert_quat(float *q) {
  fprintf(stderr, "Not implemented yet!\n");
}

/*
 * Conjugate.
 */
void conjugate_quat(float *q) {
  q[0] = -q[0];
  q[1] = -q[1];
  q[2] = -q[2];
}

/*
 * Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
 * If they don't add up to 1.0, dividing by their magnitued will
 * renormalize them.
 *
 * Note: See the following for more information on quaternions:
 * - Shoemake, K., Animating rotation with quaternion curves, Computer
 *   Graphics 19, No 3 (Proc. SIGGRAPH'85), 245-254, 1985.
 * - Pletinckx, D., Quaternion calculus as a basic tool in computer
 *   graphics, The Visual Computer 5, 2-13, 1989.
 */
void normalize_quat(float *q) {
  float mag;

  mag = sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
  if(mag) {
    float m = 1.0 / mag; 
    unsigned int i;
    for(i=0; i<4; i++) q[i] *= m;
  }
}

/*
 * Given two rotations, q1 and q2, expressed as quaternion rotations,
 * figure out the equivalent single rotation and stuff it into q.
 *
 * This routine also normalizes the result every RENORMCOUNT times it is
 * called, to keep error from creeping in.
 *
 * NOTE: This routine is written so that q1 or q2 may be the same
 * as dest (or each other).
 */
#define RENORMCOUNT 97
void add_quats(float *q, const float *q1, const float *q2) {
  static unsigned int count=0;
  float t1[4], t2[4], t3[4], tf[4];

  copy_vec3(t1, q1);
  scale_vec3(t1, q2[3]);

  copy_vec3(t2, q2);
  scale_vec3(t2, q1[3]);

  cross_vec3(t3, q2, q1);
  add_vec3(tf, t1, t2);
  add_vec3(tf, t3, tf);
  tf[3] = q1[3]*q2[3] - v3_dot(q1, q2);

  copy_vec4(q, tf);

  if (++count > RENORMCOUNT) {
    count = 0;
    normalize_quat(q);
  }
}

/*
 * Mult Quats together.
 */
void mult_quats(float *q, const float *q1, const float *q2) {
  vec3_t va, vb, vc;

  q[3] = dot_vec3(q1, q2);
  
  cross_vec3(va, q1, q2);
  scale2_vec3(vb, q1, q2[3]);
  scale2_vec3(vc, q2, q1[3]);
  add_vec3(va, va, vb);
  add_vec3(q, va, vc);

  normalize_quat(q);
}

/*
 * Sphere Linear intERPolation.
 */
#define DELTA 0.1
void slerp_quats(float *res, const float *from, const float *to, const float t) {
  float to1[4];
  double omega, cosom, sinom, scale0, scale1;

  cosom = from[1] * to[1] 
    + from[2] * to[2] 
    + from[3] * to[3]
    + from[0] * to[0];

  if(cosom <0.0) { 
    cosom = -cosom; 
    to1[0] = - to[1];
    to1[1] = - to[2];
    to1[2] = - to[3];
    to1[3] = - to[0];
  } 
  else {
    to1[0] = to[1];
    to1[1] = to[2];
    to1[2] = to[3];
    to1[3] = to[0];
  }

  if((1.0 - cosom) > DELTA) {
    omega = acos(cosom);
    sinom = sin(omega);
    scale0 = sin((1.0 - t) * omega) / sinom;
    scale1 = sin(t * omega) / sinom;
  } 
  else {        
    /* "from" and "to" quaternions are very close */
    /*  ... so we can do a linear interpolation */
    scale0 = 1.0 - t;
    scale1 = t;
  }

  res[1] = scale0 * from[1] + scale1 * to1[0];
  res[2] = scale0 * from[2] + scale1 * to1[1];
  res[3] = scale0 * from[3] + scale1 * to1[2];
  res[0] = scale0 * from[0] + scale1 * to1[3];
}

/*
 * Euler angles -> Quat.
 */
void euler_to_quat(float *quat, const float yaw, const float pitch, const float roll) {
  vec3_t vx = {1.0, 0.0, 0.0}, vy = {0.0, 1.0, 0.0}, vz = {0.0, 0.0, 1.0};
  quat_t qx, qy, qz, qt;

  axis_to_quat(qx, vx, pitch);
  axis_to_quat(qy, vy, yaw);
  axis_to_quat(qz, vz, roll);

  mult_quats(qt, qx, qy);
  mult_quats(quat, qt, qz);
}

/*
 * Given an axis and angle (in degrees), compute quaternion.
 */
void axis_to_quat(float *q, const float *a, const float phi) {
  copy_vec3(q, a);
  scale_vec3(q, m_sin(phi/2.0));

  q[3] = m_cos(phi/2.0);

  normalize_quat(q);
}

/*
 * Compute angle/axis from a quaternion.
 */
void quat_to_axis(float *a, float phi, const float *q) {
  quat_t qr;
  float c, s;

  copy_vec4(qr, q);
  normalize_quat(qr);

  c = qr[3];
  phi = m_acos(c) * 2.0;
  s = sqrt(1.0 - m_square(c));

  if(fabs(s)<FLT_EPSILON)
    s = 1.0;

  scale2_vec3(a, q, 1.0/s);
}

/* 
 * Convertion Mat -> Quat.
 */
void mat_to_quat(float *q, const float *m) {
  float tr,s;
  tr = m[0] + m[5] + m[10] + 1.0;

  if(tr > 0.0) {
    s = 0.5 / sqrt(tr);
    q[3] = 0.25 / s;
    q[0] = (m[9]-m[6]) * s;
    q[1] = (m[2]-m[8]) * s;
    q[2] = (m[4]-m[1]) * s;
  } 
  else {    
    quat_t tq;
    int i,j,k;
    int nxt[3] = {1, 2, 0};

    i = 0;
    if(m[5] > m[0]) 
      i = 1;
    if(m[10] > m[i+i*4]) 
      i = 2;
    j = nxt[i];
    k = nxt[j];

    s = sqrt(1.0 + (m[i+i*4] - (m[j+j*4] + m[k+k*4]) )) * 2.0;
                         
    tq[i] = 0.5 / s;
    tq[j] = (m[i+j*4] + m[j+i*4]) / s;
    tq[k] = (m[i+k*4] + m[k+i*4]) / s;
    tq[3] = (m[j+k*4] + m[k+j*4]) / s;

    copy_vec4(q, tq);
  }
}

/*
 * Convertion Quat -> Mat.
 */
void quat_to_mat(float *m, const float *q) {
  float xx, xy, xz, xw;
  float yy, yz, yw;
  float zz, zw;
  
  xx = m_square(q[0]);
  xy = q[0] * q[1];
  xz = q[0] * q[2];
  xw = q[0] * q[3];
  
  yy = m_square(q[1]);
  yz = q[1] * q[2];
  yw = q[1] * q[3];
  
  zz = m_square(q[2]);
  zw = q[2] * q[3];

  m[0] = 1.0 - 2.0 * (yy + zz);
  m[1] = 2.0 * (xy - zw);
  m[2] = 2.0 * (xz + yw);

  m[4] = 2.0 * (xy + zw);
  m[5] = 1.0 - 2.0 * (xx + zz);
  m[6] = 2.0 * (yz - xw);

  m[8] = 2.0 * (xz - yw);
  m[9] = 2.0 * (yz + xw);
  m[10] = 1.0 - 2.0 * (xx + yy);

  m[3] = m[7] = m[11] = m[12] = m[13] = m[14] = 0.0;
  m[15] = 1.0;
}

/*
 * Build a rotation matrix, given a quaternion rotation.
 */
void build_rotmatrix(float *m, const float *q) {
  m[0] = 1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]);
  m[4] = 2.0 * (q[0] * q[1] - q[2] * q[3]);
  m[8] = 2.0 * (q[2] * q[0] + q[1] * q[3]);
  m[12] = 0.0;

  m[1] = 2.0 * (q[0] * q[1] + q[2] * q[3]);
  m[5]= 1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]);
  m[9] = 2.0 * (q[1] * q[2] - q[0] * q[3]);
  m[13] = 0.0;

  m[2] = 2.0 * (q[2] * q[0] - q[1] * q[3]);
  m[6] = 2.0 * (q[1] * q[2] + q[0] * q[3]);
  m[10] = 1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]);
  m[14] = 0.0;

  m[3] = 0.0;
  m[7] = 0.0;
  m[11] = 0.0;
  m[15] = 1.0;
}

/*
 * Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
 * if we are away from the center of the sphere.
 */
static float tb_project_to_sphere(const float r, const float x, const float y) {
  float d, t, z;

  d = sqrt(x*x + y*y);
  if (d < r * 0.70710678118654752440) { /* inside sphere */
    z = sqrt(r*r - d*d);
  } else { /* on hyperbola */
    t = r / 1.41421356237309504880;
    z = t*t / d;
  }
  return z;
}

/*
 * Ok, simulate a track-ball.  Project the points onto the virtual
 * trackball, then figure out the axis of rotation, which is the cross
 * product of P1 P2 and O P1 (O is the center of the ball, 0,0,0)
 * Note:  This is a deformed trackball-- is a trackball in the center,
 * but is deformed into a hyperbolic sheet of rotation away from the
 * center.  This particular function was chosen after trying out
 * several variations.
 *
 * It is assumed that the arguments to this routine are in the range
 * (-1.0 ... 1.0).
 */
void trackball(float *q, 
               const float p1x, const float p1y, 
	       const float p2x, const float p2y) {
  float a[3]; /* axis of rotation */
  float phi;  /* how much to rotate about axis */
  float p1[3], p2[3], d[3];
  float t;

  if (p1x == p2x && p1y == p2y) { /* zero rotation */
    v3_zero(q);
    q[3] = 1.0;
    return;
  }

  /* First, figure out z-coordinates for projection of P1 and P2 
     to deformed sphere */
  set_vec3(p1,p1x,p1y,tb_project_to_sphere(TRACKBALLSIZE,p1x,p1y));
  set_vec3(p2,p2x,p2y,tb_project_to_sphere(TRACKBALLSIZE,p2x,p2y));

  /* Now, we want the cross product of P1 and P2 */
  cross_vec3(a,p2,p1);

  /* Figure out how much to rotate around that axis */
  sub_vec3(d,p1,p2);
  t = length_vec3(d) / (2.0*TRACKBALLSIZE);

  /* Avoid problems with out-of-control values... */
  if (t > 1.0) t = 1.0;
  if (t < -1.0) t = -1.0;
  phi = 2.0 * m_asin(t);

  axis_to_quat(q,a,phi);
}

/*
 * Test the libruary.
 */
void test_quats(void) {
  vec3_t x3,y3,z3,u3,v3,w3,t3;
  quat_t q;
  mat4_t m4,n4;

  printf("-> Testing quaternion routines:\n");

  v3_set(u3, 1.0, 0.0, 0.0);
  v3_set(v3, 0.0, 1.0, 0.0);
  printf("v3_set(u3,1.0,0.0,0.0)= <%.2f,%.2f,%.2f>\n", u3[0], u3[1], u3[2]);
  printf("v3_set(v3,0.0,1.0,0.0)= <%.2f,%.2f,%.2f>\n", v3[0], v3[1], v3[2]);
  v3_cross(w3, u3, v3);
  printf("v3_cross(w3,u3,v3)= <%.2f,%.2f,%.2f>\n", w3[0], w3[1], w3[2]);

  axis_to_quat(q, u3, 60.0);
  printf("axis_to_quat(q,u3,60.0)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
  build_rotmatrix(m4, q);
  printf("build_rotmatrix(m4,q)= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);
  quat_to_mat(n4, q);
  printf("quat_to_mat(n4,q)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);
  mat_to_quat(q, m4);
  printf("mat_to_quat(q, m4)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
  mat_to_quat(q, n4);	    
  printf("mat_to_quat(q, n4)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);

  axis_to_quat(q, v3, 60.0);
  printf("axis_to_quat(q,v3,60.0)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
  build_rotmatrix(m4, q);
  printf("build_rotmatrix(m4,q)= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);
  quat_to_mat(n4, q);
  printf("quat_to_mat(n4,q)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);
  mat_to_quat(q, m4);
  printf("mat_to_quat(q, m4)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
  mat_to_quat(q, n4);	    
  printf("mat_to_quat(q, n4)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
   
  axis_to_quat(q, w3, 60.0);
  printf("axis_to_quat(q,w3,60.0)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
  build_rotmatrix(m4, q);
  printf("build_rotmatrix(m4,q)= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("                       [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);
  quat_to_mat(n4, q);
  printf("quat_to_mat(n4,q)= [%.2f %.2f %.2f %.2f]\n", n4[0], n4[4], n4[8], n4[12]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[1], n4[5], n4[9], n4[13]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[2], n4[6], n4[10], n4[14]);
  printf("                   [%.2f %.2f %.2f %.2f]\n", n4[3], n4[7], n4[11], n4[15]);
  mat_to_quat(q, m4);
  printf("mat_to_quat(q, m4)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);
  mat_to_quat(q, n4);	    
  printf("mat_to_quat(q, n4)= <%.2f,%.2f,%.2f,%.2f>\n", q[0], q[1], q[2], q[3]);

  v3_set(t3, 10.0, 20.0, -200.0);
  m4_make(m4, t3, &m4[0], &m4[8], &m4[4]);
  v3_scale(&m4[8], -1.0);
  printf("m4_make(m4,x3,&m4[0],&m4[8],&m4[4])= [%.2f %.2f %.2f %.2f]\n", m4[0], m4[4], m4[8], m4[12]);
  printf("                                     [%.2f %.2f %.2f %.2f]\n", m4[1], m4[5], m4[9], m4[13]);
  printf("                                     [%.2f %.2f %.2f %.2f]\n", m4[2], m4[6], m4[10], m4[14]);
  printf("                                     [%.2f %.2f %.2f %.2f]\n", m4[3], m4[7], m4[11], m4[15]);

  m4_xformvec3(x3, u3, m4);
  printf("m4_xformvec3(x3,u3,m4)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  m4_xformvec3(y3, v3, m4);
  printf("m4_xformvec3(y3,v3,m4)= <%.2f,%.2f,%.2f>\n", y3[0], y3[1], y3[2]);
  m4_xformvec3(z3, w3, m4);
  printf("m4_xformvec3(z3,w3,m4)= <%.2f,%.2f,%.2f>\n", z3[0], z3[1], z3[2]);
}
