/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include <assert.h>

#include <GL/gl.h>

#include "sdlib.h"

#include "scene.h"

#include "mat.h"
#include "xml.h"

/*
 * Constructor.
 */
Scene *scene_new(void) {
  /* Allocation de la structure */
  Scene *scene = (Scene *)malloc(sizeof(Scene));
  if(!scene) {
    sd_error("Can't allocate memory (Scene)!");
  }

  /* Initialize starsfield */
  scene->stars = stars_new();

  /* Initialize space ships */
  scene->nb_space_ships = 0;

  return scene;
}

/*
 * Destructor.
 */
int scene_destroy(Scene *scene) {
  unsigned int i;

  /* Assertion */
  assert(scene);
  assert(scene->stars);
  assert(scene->space_ships[(scene->nb_space_ships)-1]);

  /* Free memory */
  stars_destroy(scene->stars);
  for(i=0; i<scene->nb_space_ships; i++) {
    space_ship_destroy(scene->space_ships[i]);
  }
  free(scene);

  return 0;
}

/*
 * Init world.
 */
int scene_parse_xml(Scene *scene, const char *scn_file) {
  Xml *scn = NULL;
  char *val = NULL;
  
  /* Assertion */
  assert(scene);
  assert(scene->stars);

  /* Load scene */
  scn = xml_new();
  printline("\nParsing scene file: %s\n", scn_file);
  xml_init(scn, scn_file);

  /* Parse xml file */
  
  /* Initialize starsfield */
  val = xml_search_path(scn, "/Stardust/scene/id/nebula/text()");
  stars_init(scene->stars, val);

  /* Initialize ships */
  {
    const char *path_player = "/Stardust/scene/player/%s/text()";
    const char *path_shp = "/Stardust/scene/ships[@*]/ship/%s/text()";
    char *path;
    unsigned int n;
    float x, y, z;

    path = malloc((strlen(path_shp) + 8) * sizeof(char));
    if(!path) {
      sd_error("Not enough memory!");
    }

    /* Loading player ship */
    scene->space_ships[0] = space_ship_new();
    scene->nb_space_ships = 1;
    sprintf(path, path_player, "shp");
    val = xml_search_path(scn, path);
    space_ship_load(scene->space_ships[0], val);
    /* Init position */
    sprintf(path, path_player, "posx");
    val = xml_search_path(scn, path);
    x = atof(val);
    sprintf(path, path_player, "posy");
    val = xml_search_path(scn, path);
    y = atof(val);
    sprintf(path, path_player, "posz");
    val = xml_search_path(scn, path);
    z = atof(val);
    set_vec3(space_ship_xyz(scene->space_ships[0]), x, y, z);
    /* Init orientation */
    /* TODO */

    /*val = xml_search_path(scn, "/Stardust/scene/ships/@nb");*/

    /* Search ships */
    n = 0; /* init path */
    sprintf(path, path_shp, "shp");
    while( (val = xml_search_xpath(scn, path, n)) ) {
      /* Load ship model */
      scene->space_ships[n+1] = space_ship_new();
      scene->nb_space_ships += 1;
      space_ship_load(scene->space_ships[n+1], val);
      /* Init position */
      sprintf(path, path_shp, "posx");
      val = xml_search_xpath(scn, path, n);
      x = atof(val);
      sprintf(path, path_shp, "posy");
      val = xml_search_xpath(scn, path, n);
      y = atof(val);
      sprintf(path, path_shp, "posz");
      val = xml_search_xpath(scn, path, n);
      z = atof(val);
      set_vec3(space_ship_xyz(scene->space_ships[n+1]), x, y, z);
      /* Init orientation */
      /* TODO */
      n++; /* go next and re-init path */
      sprintf(path, path_shp, "shp");
    }
    
    free(path);
  }

  /* Search other objects */
  /* TODO */

  /* Free memory */
  xml_destroy(scn);

  return 0;
}

/*
 * Update world.
 */
int scene_update(Scene *scene, const float dt) {
  unsigned int i;

  /* Assertion */
  assert(scene);
  assert(scene->space_ships[(scene->nb_space_ships)-1]);

  /* Update all ships */
  for(i=0; i<scene->nb_space_ships; i++) {
    space_ship_update(scene->space_ships[i], dt);
  }

  return 0;
}

/*
 * Check collisions.
 */
int scene_check_collisions(Scene *scene){
  /* Assertion */
  assert(scene);
  assert(scene->space_ships[(scene->nb_space_ships)-1]);

  return 0;
}

/*
 * Draw world.
 */
int scene_draw(Scene *scene) {
  mat4_t mat;
  unsigned int i;

  /* Assertion */
  assert(scene);
  assert(scene->stars);
  assert(scene->space_ships[(scene->nb_space_ships)-1]);

  /* Update camera in accordance with player ship */
  glLoadIdentity();
  invert_mat4(mat, scene->space_ships[0]->mat);
  glMultMatrixf(mat);

  /* Draw starsfield */
  stars_draw(scene->stars);

  /* Draw scene */
  for(i=1; i<scene->nb_space_ships; i++) {
    space_ship_draw(scene->space_ships[i]);
  }

  return 0;
}
