/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>

#include <assert.h>

#include "sdlib.h"

#include "spaceship.h"

#include "vec3d.h"
#include "mat.h"
#include "quat.h"

/*
 * Constructor.
 */
SpaceShip *space_ship_new(void) {
  SpaceShip *space_ship = (SpaceShip *)malloc(sizeof(SpaceShip));
  if(!space_ship) {
    sd_error("Can't allocate memory (SpaceShip)!");
  }
  
  /* Inits */
  zero_vec3(space_ship->xyz);
  set_vec3(space_ship->right, 1.0, 0.0, 0.0);
  set_vec3(space_ship->forward, 0.0, 0.0, 1.0);
  set_vec3(space_ship->up, 0.0, 1.0, 0.0);

  zero_vec3(space_ship->xyz_accel);
  zero_vec3(space_ship->hpr_accel);
  zero_vec3(space_ship->xyz_velocity);
  zero_vec3(space_ship->hpr_velocity);

  space_ship->model = model_desc_new();

  space_ship->assisted_mode = 1;

  space_ship->autopilot = APILOT_NONE;

  space_ship->vdu = CONSOLE_NAV;
  space_ship->hud_detail = HUD_MED_DETAIL;

  return space_ship;
}

/*
 * Destructor.
 */
int space_ship_destroy(SpaceShip *space_ship) {
  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Free memory */
  model_desc_destroy(space_ship->model);
  free(space_ship);

  return 0;
}

/*
 * Initialization.
 */
int space_ship_init(SpaceShip *space_ship) {
  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Inits */
  model_desc_init(space_ship->model);

  return 0;
}
int space_ship_load(SpaceShip *space_ship, const char *ship_file) {
  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Inits */
  model_desc_load(space_ship->model, ship_file);

  return 0;
}

/*
 * Commands.
 */
int space_ship_move_forward(SpaceShip *space_ship) {
  space_ship->xyz_accel[FORWARD] = model_desc_forward_accel(space_ship->model);
  return 0;
}
int space_ship_move_back(SpaceShip *space_ship) {
  space_ship->xyz_accel[FORWARD] = -model_desc_back_accel(space_ship->model);
  return 0;
}
int space_ship_move_left(SpaceShip *space_ship) {
  space_ship->xyz_accel[RIGHT] = -model_desc_lateral_accel(space_ship->model);
  return 0;
}
int space_ship_move_right(SpaceShip *space_ship) {
  space_ship->xyz_accel[RIGHT] = model_desc_lateral_accel(space_ship->model);
  return 0;
}
int space_ship_move_up(SpaceShip *space_ship) {
  space_ship->xyz_accel[UP] = model_desc_up_accel(space_ship->model);
  return 0;
}
int space_ship_move_down(SpaceShip *space_ship) {
  space_ship->xyz_accel[UP] = -model_desc_down_accel(space_ship->model);
  return 0;
}

int space_ship_yaw_left(SpaceShip *space_ship) {
  space_ship->hpr_accel[YAW] = -model_desc_yaw_accel(space_ship->model);
  return 0;
}
int space_ship_yaw_right(SpaceShip *space_ship) {
  space_ship->hpr_accel[YAW] = model_desc_yaw_accel(space_ship->model);
  return 0;
}
int space_ship_pitch_up(SpaceShip *space_ship) {
  space_ship->hpr_accel[PITCH] = -model_desc_pitch_accel(space_ship->model);
  return 0;
}
int space_ship_pitch_down(SpaceShip *space_ship) {
  space_ship->hpr_accel[PITCH] = model_desc_pitch_accel(space_ship->model);
  return 0;
}
int space_ship_roll_left(SpaceShip *space_ship) {
  space_ship->hpr_accel[ROLL] = -model_desc_roll_accel(space_ship->model);
  return 0;
}
int space_ship_roll_right(SpaceShip *space_ship) {
  space_ship->hpr_accel[ROLL] = model_desc_roll_accel(space_ship->model);
  return 0;
}

int space_ship_stop_move_forward(SpaceShip *space_ship) {
  space_ship->xyz_accel[FORWARD] = 0.0;
  return 0;
}
int space_ship_stop_move_lateral(SpaceShip *space_ship) {
  space_ship->xyz_accel[RIGHT] = 0.0;
  return 0;
}
int space_ship_stop_move_up_down(SpaceShip *space_ship) {
  space_ship->xyz_accel[UP] = 0.0;
  return 0;
}

int space_ship_stop_yaw(SpaceShip *space_ship) {
  space_ship->hpr_accel[YAW] = 0.0;
  return 0;
}
int space_ship_stop_pitch(SpaceShip *space_ship) {
  space_ship->hpr_accel[PITCH] = 0.0;
  return 0;
}
int space_ship_stop_roll(SpaceShip *space_ship) {
  space_ship->hpr_accel[ROLL] = 0.0;
  return 0;
}

int space_ship_to_break_off(SpaceShip *space_ship) {
  space_ship->autopilot = APILOT_TO_STOP;
  space_ship->assisted_mode = 1;

  zero_vec3(space_ship->xyz_accel);
  zero_vec3(space_ship->hpr_accel);
  /*zero_vec3(space_ship->xyz_velocity);
    zero_vec3(space_ship->hpr_velocity);*/

  space_ship->autopilot = APILOT_NONE;

  return 0;
}
int space_ship_to_come_near(SpaceShip *space_ship) {
  space_ship->autopilot = APILOT_TO_APPROACH;
  /* TODO */
  space_ship->autopilot = APILOT_NONE;

  return 0;
}
int space_ship_in_formation(SpaceShip *space_ship) {
  space_ship->autopilot = APILOT_IN_FORMATION;
  /* TODO */
  space_ship->autopilot = APILOT_NONE;

  return 0;
}
int space_ship_to_moor(SpaceShip *space_ship) {
  space_ship->autopilot = APILOT_TO_MOOR;
  /* TODO */
  space_ship->autopilot = APILOT_NONE;

  return 0;
}

/*
 * Assisted flight.
 * Return new xyz/hpr_velocity.
 */
#define SPEED_SOFTENING (1.5)
#define ROT_SOFTENING (1.0)
static int space_ship_assisted_flight(SpaceShip *space_ship, 
				      const vec3_t xyz_accel, const vec3_t hpr_accel,
				      vec3_t xyz_velocity, vec3_t hpr_velocity,
				      const float dt) {
  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Correct forward/back gliding */
  if(!xyz_accel[FORWARD]) {
    float forward_spd, back_spd;
    
    forward_spd = (model_desc_forward_accel(space_ship->model)*dt)/SPEED_SOFTENING;
    back_spd = (model_desc_back_accel(space_ship->model)*dt)/SPEED_SOFTENING;
    
    if(space_ship->xyz_velocity[FORWARD] > back_spd) {
      xyz_velocity[FORWARD] = -back_spd;
    }
    else if(space_ship->xyz_velocity[FORWARD] < -forward_spd) {
      xyz_velocity[FORWARD] = forward_spd;
    }
    else {
      xyz_velocity[FORWARD] = -space_ship->xyz_velocity[FORWARD];
    }
  }
  /* Correct lateral gliding */
  if(!xyz_accel[RIGHT]) {
    float lateral_spd;
    
    lateral_spd = (model_desc_lateral_accel(space_ship->model)*dt)/SPEED_SOFTENING;
    
    if(space_ship->xyz_velocity[RIGHT] > lateral_spd) {
      xyz_velocity[RIGHT] = -lateral_spd;
    }
    else if(space_ship->xyz_velocity[RIGHT] < -lateral_spd) {
      xyz_velocity[RIGHT] = lateral_spd;
    }
    else {
      xyz_velocity[RIGHT] = -space_ship->xyz_velocity[RIGHT];
    }
  } 
  /* Correct up/down gliding */
  if(!xyz_accel[UP]) {
    float up_spd, down_spd;
    
    up_spd = (model_desc_up_accel(space_ship->model)*dt)/SPEED_SOFTENING;
    down_spd = (model_desc_down_accel(space_ship->model)*dt)/SPEED_SOFTENING;
    
    if(space_ship->xyz_velocity[UP] > down_spd) {
      xyz_velocity[UP] = -down_spd;
    }
    else if(space_ship->xyz_velocity[UP] < -up_spd) {
      xyz_velocity[UP] = up_spd;
    }
    else {
      xyz_velocity[UP] = -space_ship->xyz_velocity[UP];
    }
  }
  
  /* Correct yaw gliding */
  if(!hpr_accel[YAW]) {
    float yaw_spd;
    
    yaw_spd = (model_desc_yaw_accel(space_ship->model)*dt)/ROT_SOFTENING;
    
    if(space_ship->hpr_velocity[YAW] > yaw_spd) {
      hpr_velocity[YAW] = -yaw_spd;
    }
    else if(space_ship->hpr_velocity[YAW] < -yaw_spd) {
      hpr_velocity[YAW] = yaw_spd;
    }
    else {
      hpr_velocity[YAW] = -space_ship->hpr_velocity[YAW];
    }
  } 
  /* Correct pitch gliding */
  if(!hpr_accel[PITCH]) {
    float pitch_spd;
    
    pitch_spd = (model_desc_pitch_accel(space_ship->model)*dt)/ROT_SOFTENING;
    
    if(space_ship->hpr_velocity[PITCH] > pitch_spd) {
      hpr_velocity[PITCH] = -pitch_spd;
    }
    else if(space_ship->hpr_velocity[PITCH] < -pitch_spd) {
      hpr_velocity[PITCH] = pitch_spd;
    }
    else {
      hpr_velocity[PITCH] = -space_ship->hpr_velocity[PITCH];
    }
  } 
  /* Correct roll gliding */
  if(!hpr_accel[ROLL]) {
    float roll_spd;
    
    roll_spd = (model_desc_roll_accel(space_ship->model)*dt)/ROT_SOFTENING;
    
    if(space_ship->hpr_velocity[ROLL] > roll_spd) {
      hpr_velocity[ROLL] = -roll_spd;
    }
    else if(space_ship->hpr_velocity[ROLL] < -roll_spd) {
      hpr_velocity[ROLL] = roll_spd;
    }
    else {
      hpr_velocity[ROLL] = -space_ship->hpr_velocity[ROLL];
    }
  }
  
  return 0;
}

/*
 * Check velocity/speed.
 */
static int space_ship_check_velocity(SpaceShip *space_ship, 
				      const vec3_t xyz_velocity, const vec3_t hpr_velocity) {
  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Check forward/back velocity */
  if(space_ship->xyz_velocity[FORWARD] > model_desc_forward_velocity(space_ship->model)) {
    space_ship->xyz_velocity[FORWARD] = model_desc_forward_velocity(space_ship->model);
  }
  else if(space_ship->xyz_velocity[FORWARD] < -model_desc_back_velocity(space_ship->model)) {
    space_ship->xyz_velocity[FORWARD] = -model_desc_back_velocity(space_ship->model);
  }
  /* Check lateral velocity */
  if(space_ship->xyz_velocity[RIGHT] > model_desc_lateral_velocity(space_ship->model)) {
    space_ship->xyz_velocity[RIGHT] = model_desc_lateral_velocity(space_ship->model);
  }
  else if(space_ship->xyz_velocity[RIGHT] < -model_desc_lateral_velocity(space_ship->model)) {
    space_ship->xyz_velocity[RIGHT] = -model_desc_lateral_velocity(space_ship->model);
  }
  /* Check up/down velocity */
  if(space_ship->xyz_velocity[UP] > model_desc_up_velocity(space_ship->model)) {
    space_ship->xyz_velocity[UP] = model_desc_up_velocity(space_ship->model);
  }
  else if(space_ship->xyz_velocity[UP] < -model_desc_down_velocity(space_ship->model)) {
    space_ship->xyz_velocity[UP] = -model_desc_down_velocity(space_ship->model);
  }

  /* Check yaw velocity */
  if(space_ship->hpr_velocity[YAW] > model_desc_yaw_velocity(space_ship->model)) {
    space_ship->hpr_velocity[YAW] = model_desc_yaw_velocity(space_ship->model);
  }
  else if(space_ship->hpr_velocity[YAW] < -model_desc_yaw_velocity(space_ship->model)) {
    space_ship->hpr_velocity[YAW] = -model_desc_yaw_velocity(space_ship->model);
  }
  /* Check pitch velocity */
  if(space_ship->hpr_velocity[PITCH] > model_desc_pitch_velocity(space_ship->model)) {
    space_ship->hpr_velocity[PITCH] = model_desc_pitch_velocity(space_ship->model);
  }
  else if(space_ship->hpr_velocity[PITCH] < -model_desc_pitch_velocity(space_ship->model)) {
    space_ship->hpr_velocity[PITCH] = -model_desc_pitch_velocity(space_ship->model);
  }
  /* Check roll velocity */
  if(space_ship->hpr_velocity[ROLL] > model_desc_roll_velocity(space_ship->model)) {
    space_ship->hpr_velocity[ROLL] = model_desc_roll_velocity(space_ship->model);
  }
  else if(space_ship->hpr_velocity[ROLL] < -model_desc_roll_velocity(space_ship->model)) {
    space_ship->hpr_velocity[ROLL] = -model_desc_roll_velocity(space_ship->model);
  }

  return 0;
}

/*
 * Update space_ship position/orientation.
 */
#define SPEED_TRICK (1.0)
#define ROT_TRICK (1.0)
int space_ship_update(SpaceShip *space_ship, const float dt) {
  vec3_t xyz_accel, hpr_accel;
  vec3_t xyz_velocity, hpr_velocity;
  vec3_t dir;
  mat4_t rmat;
  quat_t q;

  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Get acceleration and compute velocity */
  scale2_vec3(xyz_accel, space_ship->xyz_accel, SPEED_TRICK);
  scale2_vec3(hpr_accel, space_ship->hpr_accel, ROT_TRICK);
  scale2_vec3(xyz_velocity, xyz_accel, dt);
  scale2_vec3(hpr_velocity, hpr_accel, dt);

  if(space_ship->assisted_mode) {
    /* Compute assisted flight */
    space_ship_assisted_flight(space_ship, 
			       xyz_accel, hpr_accel, 
			       xyz_velocity, hpr_velocity, dt);
  }

  /* Update velocity */
  add_vec3(space_ship->xyz_velocity, space_ship->xyz_velocity, xyz_velocity);
  add_vec3(space_ship->hpr_velocity, space_ship->hpr_velocity, hpr_velocity);

  /* Check velocity */
  space_ship_check_velocity(space_ship, xyz_velocity, hpr_velocity);

  /* Translation */
  if(space_ship->xyz_velocity[FORWARD]) {
    copy_vec3(dir, space_ship->forward);
    scale_vec3(dir, space_ship->xyz_velocity[FORWARD]*dt);
    add_vec3(space_ship->xyz, space_ship->xyz, dir);
  }
  if(space_ship->xyz_velocity[RIGHT]) {
    copy_vec3(dir, space_ship->right);
    scale_vec3(dir, space_ship->xyz_velocity[RIGHT]*dt);
    add_vec3(space_ship->xyz, space_ship->xyz, dir);
  }
  if(space_ship->xyz_velocity[UP]) {
    copy_vec3(dir, space_ship->up);
    scale_vec3(dir, space_ship->xyz_velocity[UP]*dt);
    add_vec3(space_ship->xyz, space_ship->xyz, dir);
  }

  /* Rotation */
  if(space_ship->hpr_velocity[YAW]) {
    axis_to_quat(q, space_ship->up, space_ship->hpr_velocity[YAW]*dt);
    build_rotmatrix(rmat, q);
    xform_vec3(space_ship->right, space_ship->right, rmat);
    xform_vec3(space_ship->forward, space_ship->forward, rmat);
    transpose_mat4(rmat); /* upd vel vector */
    xform_vec3(space_ship->xyz_velocity, space_ship->xyz_velocity, rmat);
  }
  if(space_ship->hpr_velocity[PITCH]) {
    axis_to_quat(q, space_ship->right, space_ship->hpr_velocity[PITCH]*dt);
    build_rotmatrix(rmat, q);
    xform_vec3(space_ship->forward, space_ship->forward, rmat);
    xform_vec3(space_ship->up, space_ship->up, rmat);
    transpose_mat4(rmat); /* upd vel vector */
    xform_vec3(space_ship->xyz_velocity, space_ship->xyz_velocity, rmat);
  }
  if(space_ship->hpr_velocity[ROLL]) {
    vec3_t forward;
    scale2_vec3(forward, space_ship->forward, -1.0);
    axis_to_quat(q, forward, space_ship->hpr_velocity[ROLL]*dt);
    build_rotmatrix(rmat, q);
    xform_vec3(space_ship->right, space_ship->right, rmat);
    xform_vec3(space_ship->up, space_ship->up, rmat);
    transpose_mat4(rmat); /* upd vel vector */
    xform_vec3(space_ship->xyz_velocity, space_ship->xyz_velocity, rmat);
  }

  /* Make position/orientation matrix */
  make_mat4(space_ship->mat, space_ship->xyz, 
	    space_ship->right, space_ship->forward, space_ship->up);

  return 0;
}

/*
 * Draw spaceship at position/orientation.
 */
int space_ship_draw(SpaceShip *space_ship) {
  /* Assertion */
  assert(space_ship);
  assert(space_ship->model);

  /* Draw space ship */
  glPushMatrix();
  glMultMatrixf(space_ship->mat);
  model_desc_draw(space_ship->model);
  glPopMatrix();

  return 0;
}
