/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Definition of the SpaceShip struct.,
 * and public interface.
 */

#ifndef _SPACESHIP_H
#define _SPACESHIP_H

#include <GL/gl.h>

#include "vec3d.h"

#include "modeldesc.h"

/* Session type */
#define RIGHT   0
#define FORWARD 2
#define UP      1
#define YAW   1
#define PITCH 0
#define ROLL  2

/* Automatic pilot */
#define APILOT_NONE 0
#define APILOT_TO_STOP 1
#define APILOT_TO_APPROACH 2
#define APILOT_IN_FORMATION 3
#define APILOT_TO_MOOR 4

/* Visual Display Unit */
#define CONSOLE_1 1
#define CONSOLE_NAV CONSOLE_1
#define CONSOLE_2 2
#define CONSOLE_ARM CONSOLE_2
#define CONSOLE_3 3
#define CONSOLE_NRJ CONSOLE_3
#define CONSOLE_4 4
#define CONSOLE_CMD CONSOLE_4

/* HUD details */
#define HUD_NO_DETAIL 0
#define HUD_LOW_DETAIL 1
#define HUD_MED_DETAIL 2
#define HUD_HIGH_DETAIL 3

/* Camera mode */
#define OUTER_VIEW 0
#define REAR_VIEW 1
#define COCKPIT_VIEW 2
#define FREE_VIEW 3

/*
 * The SpaceShip.
 */
typedef struct _SpaceShip {
  vec3_t xyz; /* position */
  vec3_t right, forward, up; /* orientation */
  
  GLfloat mat[16]; /* transform */

  vec3_t xyz_accel;
  vec3_t hpr_accel;

  vec3_t xyz_velocity;
  vec3_t hpr_velocity;

  ModelDesc *model; /* model description */

  short assisted_mode;

  short autopilot; /* autopilot mode */

  short vdu; /* visual display unit */
  short hud_detail; /* level of details */

} SpaceShip;

/*
 * Constructor.
 */
SpaceShip *space_ship_new(void);

/*
 * Destructor.
 */
int space_ship_destroy(SpaceShip *space_ship);

/*
 * Access.
 */
#define space_ship_xyz(space_ship) ((space_ship)->xyz)
#define space_ship_right(space_ship) ((space_ship)->right)
#define space_ship_forward(space_ship) ((space_ship)->forward)
#define space_ship_up(space_ship) ((space_ship)->up)

#define space_ship_mat(space_ship) ((space_ship)->mat)

#define space_ship_xyz_velocity(space_ship) ((space_ship)->xyz_velocity)
#define space_ship_speed(space_ship) (length_vec3((space_ship)->xyz_velocity))

#define space_ship_model(space_ship) ((space_ship)->model)

#define space_ship_assisted_mode(space_ship) ((space_ship)->assisted_mode)
#define space_ship_swap_assisted_mode(space_ship) ((space_ship)->assisted_mode = \
                                                   1 - (space_ship)->assisted_mode)

#define space_ship_autopilot(space_ship) ((space_ship)->autopilot)

#define space_ship_visual_display_unit(space_ship) ((space_ship)->vdu)
#define space_ship_vdu space_ship_visual_display_unit
#define space_ship_vdu_nav(space_ship) ((space_ship)->vdu = CONSOLE_NAV)
#define space_ship_vdu_arm(space_ship) ((space_ship)->vdu = CONSOLE_ARM)
#define space_ship_vdu_nrj(space_ship) ((space_ship)->vdu = CONSOLE_NRJ)
#define space_ship_vdu_cmd(space_ship) ((space_ship)->vdu = CONSOLE_CMD)

#define space_ship_hud_detail(space_ship) ((space_ship)->hud_detail)
#define space_ship_change_hud(space_ship) ((space_ship)->hud_detail = \
                                           ((space_ship)->hud_detail+1)%(HUD_HIGH_DETAIL+1))

#define space_ship_view(space_ship) (OUTER_VIEW)

/*
 * Initialization.
 */
int space_ship_init(SpaceShip *space_ship);
int space_ship_load(SpaceShip *space_ship, const char *ship_file);

/*
 * Commands.
 */
int space_ship_move_forward(SpaceShip *space_ship);
#define space_ship_accelerate space_ship_move_forward 
int space_ship_move_back(SpaceShip *space_ship);
#define space_ship_brake space_ship_move_back
#define space_ship_decelerate space_ship_move_back
int space_ship_move_left(SpaceShip *space_ship);
int space_ship_move_right(SpaceShip *space_ship);
int space_ship_move_up(SpaceShip *space_ship);
int space_ship_move_down(SpaceShip *space_ship);

int space_ship_yaw_left(SpaceShip *space_ship);
int space_ship_yaw_right(SpaceShip *space_ship);
int space_ship_pitch_up(SpaceShip *space_ship);
int space_ship_pitch_down(SpaceShip *space_ship);
int space_ship_roll_left(SpaceShip *space_ship);
int space_ship_roll_right(SpaceShip *space_ship);

int space_ship_stop_move_forward(SpaceShip *space_ship);
int space_ship_stop_move_lateral(SpaceShip *space_ship);
int space_ship_stop_move_up_down(SpaceShip *space_ship);

int space_ship_stop_yaw(SpaceShip *space_ship);
int space_ship_stop_pitch(SpaceShip *space_ship);
int space_ship_stop_roll(SpaceShip *space_ship);

int space_ship_to_break_off(SpaceShip *space_ship);
int space_ship_to_come_near(SpaceShip *space_ship);
int space_ship_in_formation(SpaceShip *space_ship);
int space_ship_to_moor(SpaceShip *space_ship);

/*
 * Update spaceship position/orientation.
 */
int space_ship_update(SpaceShip *space_ship, const float dt);

/*
 * Draw spaceship at position/orientation.
 */
int space_ship_draw(SpaceShip *space_ship);


#endif /* _SPACESHIP_H */
