/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>

#include <assert.h>

#include <GL/glu.h>

#include "sdlib.h"

#include "stars.h"

/*
 * Constructor.
 */
Stars *stars_new(void) {
  /* Allocation de la structure */
  Stars *stars = (Stars *)malloc(sizeof(Stars));
  if(!stars) {
    sd_error("Can't allocate memory (Stars)!");
  }
  
  /* Inits */
  stars->id = 0;
  stars->nebula = texture_new();
  
  return stars;
}

/*
 * Destructor.
 */
int stars_destroy(Stars *stars) {
  /* Assertion */
  assert(stars);
  assert(stars->nebula);

  /* Free memory */
  texture_destroy(stars->nebula);
  free(stars);

  return 0;
}

/* 
 * Initializes an array of stars, or points, each of which    
 * are unit length of one away from (0,0,0). 
 * When nebula (texture file name) is not NULL, load the texture.
 * It then compiles the starfield into a display list.                         
 */
GLuint stars_init(Stars *stars, const char *nebula_file) {
  /* Assertion */
  assert(stars);
  assert(stars->nebula);

  /* Keyword to display only GL_POINTS */
  if(!strcmp(nebula_file, "GL_POINTS")) { 
    unsigned int n;

    /* Initialize an array of stars (GL_POINTS) */
    srandom(10);
#if 0
    {
      GLfloat x,y,z,r;
      GLfloat t;

      for(n=0; n<NUMSTARS; n++) {
	z = (float)n;
	z -= NUMSTARS/2.0;
	r = sqrt((NUMSTARS/2.0)*(NUMSTARS/2.0)-z*z);
	t = (float)(random())/(float)(RAND_MAX);
	t *= 2*M_PI;
	x = r * cos(t);
	y = r * sin(t);
	set_vec3(stars->xyz[n], x, y, z);
      }
    }
#else
    {
      GLfloat x,y,z,r;
      GLfloat h,p;
      
      for(n=0; n<NUMSTARS; n++) {
	r = 5.0; /*45000.0;*/
	h = 2*M_PI * (float)(random())/(float)(RAND_MAX);
	p = 2*M_PI * (float)(random())/(float)(RAND_MAX);
	x = r * sin(p) * sin(h);
	y = r * cos(p);
	z = r * sin(p) * cos(h);
	set_vec3(stars->xyz[n], x, y, z);
      }
    }
#endif

    /* Compile the DisplayList */
    stars->id = glGenLists(1);
    glNewList(stars->id, GL_COMPILE);

    glDisable(GL_LIGHTING);
    glDepthMask(GL_FALSE);
    glCullFace(GL_FRONT);

    /*glEnable(GL_POINT_SMOOTH);
      glPointSize(1);*/

    /*glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);*/

    glColor3f(1.0, 1.0, 1.0);
    glBegin(GL_POINTS);
    for(n=0; n<NUMSTARS; n+=3)
      glVertex3fv(stars->xyz[n]);
    glEnd();

    glColor3f(1.0, 0.7, 0.7);
    glBegin(GL_POINTS);
    for(n=1; n<NUMSTARS; n+=3)
      glVertex3fv(stars->xyz[n]);
    glEnd();

    glColor3f(0.6, 0.6, 1.0);
    glBegin(GL_POINTS);
    for(n=2; n<NUMSTARS; n+=3)
      glVertex3fv(stars->xyz[n]);
    glEnd();

    glEnable(GL_LIGHTING);
    glDepthMask(GL_TRUE);
    glCullFace(GL_BACK);

    glEndList();

  }
  else {
    unsigned int n;
    GLUquadricObj *q = NULL;
    GLuint nebula_texid;

    /* Load nebula texture file */
    printline("\nLoading nebula texture: %s\n", nebula_file);
    nebula_texid = texture_init(stars->nebula, nebula_file);

    /* Initialize an array of stars (GL_POINTS) */
    srandom(10);
#if 1
    {
      GLfloat x,y,z,r;
      GLfloat t;

      for(n=0; n<NBSTARS; n++) {
	z = (float)n;
	z -= NBSTARS/2.0;
	r = sqrt((NBSTARS/2.0)*(NBSTARS/2.0)-z*z);
	t = (float)(random())/(float)(RAND_MAX);
	t *= 2*M_PI;
	x = r * cos(t);
	y = r * sin(t);
	set_vec3(stars->xyz[n], x, y, z);
      }
    }
#else
    {
      GLfloat x,y,z,r;
      GLfloat h,p;
      
      for(n=0; n<NBSTARS; n++) {
	r = 4.5; /*45000.0;*/
	h = 2*M_PI * (float)(random())/(float)(RAND_MAX);
	p = 2*M_PI * (float)(random())/(float)(RAND_MAX);
	x = r * sin(p) * sin(h);
	y = r * cos(p);
	z = r * sin(p) * cos(h);
	set_vec3(stars->xyz[n], x, y, z);
      }
    }
#endif

    /* Compile the DisplayList */
    stars->id = glGenLists(1);
    glNewList(stars->id, GL_COMPILE);

    glDisable(GL_LIGHTING);
    glDepthMask(GL_FALSE);
    glCullFace(GL_FRONT);
    glEnable(GL_TEXTURE_2D);
    
    /* Display nebula upon sphere */
    glColor3f(1.0, 1.0, 1.0);
    glBindTexture(GL_TEXTURE_2D, nebula_texid);

    q = gluNewQuadric();
    gluQuadricDrawStyle(q, GLU_FILL);
    gluQuadricNormals(q, GLU_SMOOTH);
    gluQuadricTexture(q, GLU_TRUE);
    gluQuadricOrientation(q, GLU_INSIDE);
    glPushMatrix();
    glRotatef(-90.0, 1.0, 0.0, 0.0);
    gluSphere(q, 5.0, 16, 16);
    glPopMatrix();
    gluDeleteQuadric(q);

    glDisable(GL_TEXTURE_2D);

#if 1
    /* Add now some GL_POINTS */
    glColor3f(1.0, 1.0, 1.0);
    glBegin(GL_POINTS);
    for(n=0; n<NBSTARS; n+=3)
      glVertex3fv(stars->xyz[n]);
    glEnd();

    glColor3f(1.0, 0.7, 0.7);
    glBegin(GL_POINTS);
    for(n=1; n<NBSTARS; n+=3)
      glVertex3fv(stars->xyz[n]);
    glEnd();

    glColor3f(0.6, 0.6, 1.0);
    glBegin(GL_POINTS);
    for(n=2; n<NBSTARS; n+=3)
      glVertex3fv(stars->xyz[n]);
    glEnd();
#endif

    glEnable(GL_LIGHTING);
    glDepthMask(GL_TRUE);
    glCullFace(GL_BACK);

    glEndList();
  }

  return stars->id;
}

/*
 * Draws a starsfield by changing the modelview matrix   
 * to (0,0,0) but preserving the orientation.           
 * We then draw all the stars, which were precalculated 
 * to all be a unit length of 1 away.                   
 */
int stars_draw(Stars *stars) {
  GLfloat m[16];

  /* Assertion */
  assert(stars);

  glPushMatrix();

  /* Remove tranlation */
  glGetFloatv(GL_MODELVIEW_MATRIX, m);
  m[12] = m[13] = m[14] = 0;
  glLoadMatrixf(m);
  /* Draw */
  glCallList(stars->id);

  glPopMatrix();

  return 0;
}
