/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <assert.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <GL/glu.h>

#ifdef HAVE_LIBTIFF
#include <tiffio.h>
#endif

#include "sdlib.h"

#include "texture.h"

/*
 * Constructor.
 */
Texture *texture_new(void) {
  /* Allocation de la structure */
  Texture *texture = (Texture *)malloc(sizeof(Texture));
  if(!texture) {
    sd_error("Can't allocate memory (Texture)!");
  }

  /* Texture init */
  texture->texid = 0;
  texture->width  = 0;
  texture->height = 0;
  texture->spp    = 0;

  return texture;
}

/*
 * Destructor.
 */
int texture_destroy(Texture *texture) {
  /* Assertion */
  assert(texture);

  /* Free memory */
  free(texture);

  return 0;
}

/* 
 * Create a default checkered texture.
 */
static GLuint make_checkered_texture(Texture *texture) {
  GLuint texid = 0;
  GLubyte image[64][64][3];
  unsigned int i, j, r, g;

  /* Assertion */
  assert(texture);

  /* Default size */
  texture->width  = 64;
  texture->height = 64;
  texture->spp    = 3;

  /* Checkered texture (green, red) */
  for(i=0; i<64; i++) {
    for(j=0; j<64; j++) {
      r = ( ((i&0x8)==0)^((j&0x8)==0) ) * 255;
      g = 255 - r;
      image[i][j][0] = (GLubyte)r;
      image[i][j][1] = (GLubyte)g;
      image[i][j][2] = 0;
    }
  }

  /* Texture creation */
  glGenTextures(1, &texid);
  glBindTexture(GL_TEXTURE_2D, texid);
#if 0
  glTexImage2D(GL_TEXTURE_2D, 0, 3, 64, 64, 0,
	       GL_RGB, GL_UNSIGNED_BYTE, &image[0][0][0]);
#else
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 64, 64,
                    GL_RGB, GL_UNSIGNED_BYTE, &image[0][0][0]);
#endif

  texture->texid = texid;
  return texid;
}

/*
 * Load Truevision Targa image.
 * Return the texture id (for glBindTexture).
 */
static GLuint tga_load(Texture *texture, const char *tga_file) {
  /* TGA file */
  unsigned char image_type_code;
  short int image_width;
  short int image_height;
  unsigned char bit_count;
  unsigned char *image_data;
  /* Local variables */
  FILE *file;
  unsigned char uchar_bad;
  short int sint_bad;
  long image_size;
  int color_mode; /* 3 for RGB, 4 for RGBA */
  long image_idx;
  unsigned char color_swap;

  /* Open the TGA file */
  printline("Loading texture (%s) ...\n", tga_file);
  file = fopen(tga_file, "rb");
  if(!file) {
    char tga_file2[FILENAME_LGTH];

    dprintf("Trying in directory: "DATA_DIR"\n");
    strcpy(tga_file2, DATA_DIR);
    strcat(tga_file2, tga_file);

    file = fopen(tga_file2, "rb");
    if(!file) {
      sd_warning("Can't open file %s!", tga_file);
      sd_warning("Make a default checkered texture!");
      return make_checkered_texture(texture);
    }
  }
  
  /* Read first two bytes of data we don't need */
  fread(&uchar_bad, sizeof(unsigned char), 1, file);
  fread(&uchar_bad, sizeof(unsigned char), 1, file);

  /* Read in the image type */
  fread(&image_type_code, sizeof(unsigned char), 1, file);

  /* For our purpose, the image type should be either a 2 or a 3 */
  if((image_type_code!=2)&&(image_type_code!=3)) {
    fclose(file);
    sd_warning("Only uncompressed RGB or black-and-withe images supported!");
    sd_warning("Make a default checkered texture!");
    return make_checkered_texture(texture);
  }

  /* Read 13 bytes of data we don't need */
  fread(&sint_bad, sizeof(short int), 1, file);
  fread(&sint_bad, sizeof(short int), 1, file);
  fread(&uchar_bad, sizeof(unsigned char), 1, file);
  fread(&sint_bad, sizeof(short int), 1, file);
  fread(&sint_bad, sizeof(short int), 1, file);

  /* Read image dimensions and depth */
  fread(&image_width, sizeof(short int), 1, file);
  fread(&image_height, sizeof(short int), 1, file);
  printline(" Image: %u x %u\n", image_width, image_height);
  fread(&bit_count, sizeof(unsigned char), 1, file);
  printline(" Bit count: %d\n", bit_count);

  /* Read 1 byte of data we don't need */
  fread(&uchar_bad, sizeof(unsigned char), 1, file);

  /* Color mode -> 3 = BGR, 4 = BGRA */
  color_mode = bit_count / 8;
  image_size = image_width * image_height * color_mode;

  /* Read in image data */
  image_data = (unsigned char *)malloc(image_size*sizeof(unsigned char));
  fread(image_data, sizeof(unsigned char), image_size, file);

  /* Change BGR to RGB so OpenGL can read the image data */
  for(image_idx=0; image_idx < image_size; image_idx += color_mode) {
    color_swap = image_data[image_idx];
    image_data[image_idx] = image_data[image_idx+2];
    image_data[image_idx+2] = color_swap;
  }

  fclose(file);

  /* Texture creation */
  texture->spp = color_mode;
  texture->width = image_width;
  texture->height = image_height;

  glGenTextures(1, &(texture->texid));
  glBindTexture(GL_TEXTURE_2D, texture->texid);
  gluBuild2DMipmaps(GL_TEXTURE_2D, color_mode, image_width, image_height,
		    (color_mode==4)? GL_RGBA : GL_RGB, 
		    GL_UNSIGNED_BYTE, image_data);

  return texture->texid;
}

#ifdef HAVE_LIBTIFF
/* 
 * Read tiff texture infos (tags).
 */ 
static int tiff_read_tags(TIFF *tif) {
  char *artist = NULL, *desc = NULL;
  uint32 width  = 0;
  uint32 length = 0;
  uint16 compression = 0;
  uint32 bps = 0;
  uint16 spp = 0;
      
  /* Assertion */
  assert(tif);

  /* Artist, description */
  TIFFGetField(tif, TIFFTAG_ARTIST, &artist);
  TIFFGetField(tif, TIFFTAG_IMAGEDESCRIPTION, &desc);
  printline(" Artist: %s %s\n", artist, desc);
  
  /* Image properties */
  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &length);
  printline(" Image: %lu x %lu\n", width, length);
  
  /* Image compression algorithm */
  TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression);
  printline(" Compression %u: ", compression);
  switch(compression) {
  case 1:
    printline("none\n");
    break;
  case 2:
    printline("CCITT 1D Huffman\n");
    break;
  case 3:
    printline("CCITT 3 Facsimile\n");
    break;
  case 4:
    printline("CCITT Group 4 Facsimile\n");
    break;
  case 5:
    printline("Lempel-Ziv & Welch (Unisys patent!)\n");
    break;
  case 7:
    {
      int jpegquality = -1;
      TIFFGetField(tif, TIFFTAG_JPEGQUALITY, &jpegquality);
      printline("baseline JPEG %d\n", jpegquality);
    }
    break;
  case 32771:
    printline("word-aligned 1D Huffman\n");
    break;
  case 32773:
    printline("PackBits (Mac)\n");
    break;
  case 32809:
    printline("4-bit ThunderScan (decompression only)\n");
    break;
  case 32766:
    printline("NeXT's 2-bit (decompression only)\n");
    break;
  case 32946:
    {
      int zipquality = -1;
      TIFFGetField(tif, TIFFTAG_ZIPQUALITY, &zipquality);
      printline("Deflate zlib (LZ-style algorithm) %d\n", zipquality);
    }
    break;
  case 32845:
    printline("CIE LogLuv\n");
    break;
  default:
    printline("unknown\n");
  }

  /* Image bpp */
  TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bps);
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
  printline(" Bits per sample: %lu\n", bps);
  printline(" Samples per pixel: %u\n", spp);

  return 0;
}

/*
 * Load tiff image (max 512x512).
 * Return the texture id (for glBindTexture).
 */
static GLuint tiff_load(Texture *texture, TIFF* tif) {
  GLuint texid = 0;
  uint32 w, l;
  uint16 spp = 0;
  size_t npixels;
  uint32 *raster;

  /* Assertion */
  assert(texture);
  assert(tif);

  /* Get texture properties */
  w = texture->width;
  l = texture->height;
  spp = texture->spp;

  /* Check texture props */
  if(spp != 4 && spp != 3) {
    sd_warning("Error %d samples per pixel not supported yet!", spp);
    sd_warning("Make a default checkered texture!");
    return make_checkered_texture(texture);
  }
  if( ( (w != 512) && (w != 256) && (w != 128) && (w != 64) ) || 
      ( (l != 512) && (l != 256) && (l != 128) && (l != 64) ) ) {
    sd_warning("Error image must be 2^m x 2^n (64x64..512x512)!");
    sd_warning("Build a default checkered texture!");
    return make_checkered_texture(texture);
  }

  npixels = w * l;
  raster = (uint32 *)_TIFFmalloc(npixels*sizeof(uint32));
  if(raster != NULL) {
    /* Read the image */
    if(TIFFReadRGBAImage(tif, w, l, raster, 1)) {
      unsigned int i, j, k;
      GLubyte image[l][w][spp];

      /* Copy image */
      for(i=0; i<l; i++) {
	for(j=0; j<w; j++) {
	  for(k=0; k<spp; k++) {
	    image[i][j][k] = ((GLubyte *)raster)[i*w*4+j*4+k];
	  }
	}
      }
      
      /* Texture creation */
      glGenTextures(1, &texid);
      glBindTexture(GL_TEXTURE_2D, texid);
      gluBuild2DMipmaps(GL_TEXTURE_2D, spp, w, l,
			(spp==4)? GL_RGBA : GL_RGB, 
			GL_UNSIGNED_BYTE, &image[0][0][0]);
      
    }
    else {
      sd_error("Error while loading file!");
    }
    _TIFFfree(raster);
  }
  else {
    sd_error("Can't allocate memory (raster)!");
  }

  return texid;
}

/*
 * Load Tagged Image File Format.
 * Return the texture id (for glBindTexture).
 */
static GLuint tiff_read(Texture *texture, const char* tiff_file) {
  TIFF *tif = NULL;

  /* Assertion */
  assert(texture);

  /* Open the TIFF file */
  printline("Loading texture (%s) ...\n", tiff_file);
  tif = TIFFOpen(tiff_file, "r");
  if(!tif) {
    char tiff_file2[FILENAME_LGTH];

    printline("Trying in directory: "DATA_DIR"\n");
    strcpy(tiff_file2, DATA_DIR);
    strcat(tiff_file2, tiff_file);

    tif = TIFFOpen(tiff_file2, "r");
    if(!tif) {
      sd_warning("Can't open file %s!", tiff_file);
      sd_warning("Use a default checkered texture!");
      return make_checkered_texture(texture);
    }
  }

  { /* else read tiff image */
    uint32 l, w;
    uint16 spp = 0;

    /* Print some image infos */
    tiff_read_tags(tif);

    /* Get/store image width, height */
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &l);
    texture->width  = w;
    texture->height = l;

    /* Get/store format (samples per pixel) */
    TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
    texture->spp = spp;

    /* Load/create texture */
    texture->texid = tiff_load(texture, tif);
    
    TIFFClose(tif);
  }

  return texture->texid;
}
#endif /* HAVE_LIBTIFF */

/*
 * Init texture.
 * Return the texture id (for glBindTexture).
 */
GLuint texture_init(Texture *texture, const char *tex_file) {
  /* Assertion */
  assert(texture);

  if(strstr(tex_file, ".tga")) {
    return tga_load(texture, tex_file);
  }
  else if (strstr(tex_file, ".tif") || strstr(tex_file, ".tiff")) {
#ifdef HAVE_LIBTIFF
    return tiff_read(texture, tex_file);
#else
    sd_warning("Lib TIFF not present!");
    return make_checkered_texture(texture);
#endif
  }
  else {
    sd_warning("Image format neither TGA nor TIFF!");
    return make_checkered_texture(texture);
  }
}
