/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "vec3d.h"

/*
 * Make null vector.
 */
void zero_vec2(float *v) {
  v[0] = 0.0;
  v[1] = 0.0;
}
void zero_vec3(float *v) {
  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;
}
void zero_vec4(float *v) {
  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;
  v[3] = 0.0;
}

/*
 * Set vector.
 */
void set_vec2(float *v, const float x, const float y) {
  v[0] = x;
  v[1] = y;
}
void set_vec3(float *v, const float x, const float y, const float z) {
  v[0] = x;
  v[1] = y;
  v[2] = z;
}
void set_vec4(float *v, const float x, const float y, const float z, const float w) {
  v[0] = x;
  v[1] = y;
  v[2] = z;
  v[3] = w;
}

/* 
 * Copy src in dst.
 */
void copy_vec2(float *dst, const float *src) {
  dst[0] = src[0];
  dst[1] = src[1];
}
void copy_vec3(float *dst, const float *src) {
  dst[0] = src[0];
  dst[1] = src[1];
  dst[2] = src[2];
}
void copy_vec4(float *dst, const float *src) {
  dst[0] = src[0];
  dst[1] = src[1];
  dst[2] = src[2];
  dst[3] = src[3];
}

/* 
 * Addition : dst = a + b .
 */
void add_vec3(float *dst, const float *a, const float *b) {
  dst[0] = a[0] + b[0];
  dst[1] = a[1] + b[1];
  dst[2] = a[2] + b[2];
}

/*
 * Difference : dst = a - b .
 */
void sub_vec3(float *dst, const float *a, const float *b) {
  dst[0] = a[0] - b[0];
  dst[1] = a[1] - b[1];
  dst[2] = a[2] - b[2];
}

/*
 * Minus.
 */
void negate_vec3(float *v) {
  v[0] = -v[0];
  v[1] = -v[1];
  v[2] = -v[2];
}

/* 
 * Scale : dst = scale * src .
 */
void scale_vec3(float *v, const float div) {
  v[0] *= div;
  v[1] *= div;
  v[2] *= div;
}
void scale2_vec3(float *dst, const float *src, const float scale) {
  dst[0] = src[0] * scale;
  dst[1] = src[1] * scale;
  dst[2] = src[2] * scale;
}
void scale2_vec4(float *dst, const float *src, const float scale) {
  dst[0] = src[0] * scale;
  dst[1] = src[1] * scale;
  dst[2] = src[2] * scale;
  dst[4] = src[4] * scale;
}

/*
 * Length.
 */
float length_vec3(const float *v) {
  return sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
}

/*
 * Normalize.
 */
void normalize_vec3(float *v) {
  if(length_vec3(v))
    scale_vec3(v,1.0/length_vec3(v));
}
void norm_vec3(float *vec) {
  float len;

  len = sqrt(vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2]);
  if(len) {
    float scale = 1.0/len;
    vec[0] *= scale;
    vec[1] *= scale;
    vec[2] *= scale;
  }
}

/*
 * Dot prod : a . b .
 */
float dot_vec3(const float *a, const float *b) {
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

/* 
 * Cross prod : cross = v1 ^ v2 .
 */
void cross_vec3(float *cross, const float *v1, const float *v2) {
  float tmp[3];

  tmp[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
  tmp[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
  tmp[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
  copy_vec3(cross, tmp);
}

/*
 * Compute Euler angles from direction vector (with zero roll).
 */
void hpr_from_vec3(float *hpr, const float *dir) {
  float tmp[3];

  copy_vec3(tmp, dir);
  normalise_vec3(tmp);
  /*hpr[0] = - m_atan2(tmp[0], tmp[1]);
    hpr[1] = - m_atan2(tmp[2], sqrt(m_square(tmp[0]) + m_square(tmp[1])));*/
  hpr[0] = m_atan2(tmp[0], tmp[2]);
  hpr[1] = - m_atan2(tmp[1], sqrt(m_square(tmp[0]) + m_square(tmp[2])));
  hpr[2] = 0.0;
}

/*
 * Linear intERPolation.
 */
void lerp_vec3(float *v, const float *from, const float *to, const float t) {
  v[0] = from[0] + t * (to[0] - from[0]);
  v[1] = from[1] + t * (to[1] - from[1]);
  v[2] = from[2] + t * (to[2] - from[2]);
}

/*
 * Distance between a and b.
 */
float dist_vec3(const float *a, const float *b) {
  return sqrt(m_square(b[0]-a[0]) + m_square(b[1]-a[1]) + m_square(b[2]-a[2]));
}

/*
 * Test the libruary.
 */
void test_vectors(void) {
  vec3_t u3,v3,w3,t3;
  vec3_t x3,y3,z3;
  vec3_t hpr3;

  printf("-> Testing vector routines:\n");
   
  v3_zero(u3);
  printf("v3_zero(u3)= <%.2f,%.2f,%.2f>\n", u3[0], u3[1], u3[2]);

  v3_set(t3, 1.0, 2.0, 3.0);
  printf("v3_set(t3,1.0,2.0,3.0)= <%.2f,%.2f,%.2f>\n", t3[0], t3[1], t3[2]);

  v3_set(u3, 1.0, 0.0, 0.0);
  v3_set(v3, 0.0, 1.0, 0.0);
  printf("v3_set(u3,1.0,0.0,0.0)= <%.2f,%.2f,%.2f>\n", u3[0], u3[1], u3[2]);
  printf("v3_set(v3,0.0,1.0,0.0)= <%.2f,%.2f,%.2f>\n", v3[0], v3[1], v3[2]);

  v3_cross(w3, u3, v3);
  printf("v3_cross(w3,u3,v3)= <%.2f,%.2f,%.2f>\n", w3[0], w3[1], w3[2]);

  v3_copy(x3, u3);
  v3_copy(y3, v3);
  v3_copy(z3, w3);
  printf("v3_copy(x3,u3)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  printf("v3_copy(y3,v3)= <%.2f,%.2f,%.2f>\n", y3[0], y3[1], y3[2]);
  printf("v3_copy(z3,w3)= <%.2f,%.2f,%.2f>\n", z3[0], z3[1], z3[2]);

  v3_add(x3, u3, t3);
  v3_add(y3, v3, t3);
  v3_add(z3, w3, t3);
  printf("v3_add(x3,u3,t3)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  printf("v3_add(y3,v3,t3)= <%.2f,%.2f,%.2f>\n", y3[0], y3[1], y3[2]);
  printf("v3_add(z3,w3,t3)= <%.2f,%.2f,%.2f>\n", z3[0], z3[1], z3[2]);

  v3_sub(x3, x3, t3);
  v3_sub(y3, y3, t3);
  v3_sub(z3, z3, t3);
  printf("v3_sub(x3,x3,t3)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);
  printf("v3_sub(y3,y3,t3)= <%.2f,%.2f,%.2f>\n", y3[0], y3[1], y3[2]);
  printf("v3_sub(z3,z3,t3)= <%.2f,%.2f,%.2f>\n", z3[0], z3[1], z3[2]);

  v3_negate(x3);
  printf("v3_negate(x3)= <%.2f,%.2f,%.2f>\n", x3[0], x3[1], x3[2]);

  v3_scale(y3, 0.5);
  printf("v3_scale(y3,0.5)= <%.2f,%.2f,%.2f>\n", y3[0], y3[1], y3[2]);

  v3_scale2(z3, z3, 0.3);
  printf("v3_scale2(z3,z3,0.3)= <%.2f,%.2f,%.2f>\n", z3[0], z3[1], z3[2]);

  printf("v3_length(t3)= %.2f\n", v3_length(t3));

  v3_norm(t3);
  printf("v3_norm(t3)= <%.2f,%.2f,%.2f>\n", t3[0], t3[1], t3[2]);

  printf("v3_dot(x3,t3)= %.2f\n", v3_dot(x3, t3));

  v3_cross(w3, v3, u3);
  printf("v3_cross(w3,v3,u3)= <%.2f,%.2f,%.2f>\n", w3[0], w3[1], w3[2]);

  hpr_from_vec3(hpr3, u3);
  printf("hpr_from_vec3(hpr3,u3)= <%.2f,%.2f,%.2f>\n", 
	 hpr3[0], hpr3[1], hpr3[2]);

  hpr_from_vec3(hpr3, v3);	  
  printf("hpr_from_vec3(hpr3,v3)= <%.2f,%.2f,%.2f>\n", 
	 hpr3[0], hpr3[1], hpr3[2]);

  hpr_from_vec3(hpr3, w3);	  
  printf("hpr_from_vec3(hpr3,w3)= <%.2f,%.2f,%.2f>\n", 
	 hpr3[0], hpr3[1], hpr3[2]);

  lerp_vec3(w3, u3, v3, 0.5);
  printf("lerp_vec3(w3,u3,v3,0.5)= <%.2f,%.2f,%.2f>\n", 
	 w3[0], w3[1], w3[2]);

  lerp_vec3(w3, u3, v3, 0.6);
  printf("lerp_vec3(w3,u3,v3,0.6)= <%.2f,%.2f,%.2f>\n", 
	 w3[0], w3[1], w3[2]);

  printf("v3_dist(x3,t3)= %.2f\n", v3_dist(x3, t3));
}
