/* 
 * Stardust :-) 
 * A Free Space Flight Simulator for a Free Universe!
 * Copyright (C) 2003 MiKael NAVARRO
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* 
 * Simple vector routines. This file is meant to be included 
 * in a .c file, hence all the routines are defined statically.
 * A good C compiler should be able to automatically inline all
 * these routines. 
 */

#ifndef _VEC3D_H
#define _VEC3D_H

#include <math.h>

/* Some <math.h> files do not define M_PI... */
#ifndef M_PI
#define M_PI 3.14159265359f
#endif

#define FLT_EPSILON 1.19209290e-07f /* 0.0005 */
#define DEG_TO_RAD  (M_PI/180.0)
#define RAD_TO_DEG  (180.0/M_PI)

/* Some math macros */
#define m_sin(x) ( sin((x) * DEG_TO_RAD) )
#define m_cos(x) ( cos((x) * DEG_TO_RAD) )
#define m_tan(x) ( tan((x) * DEG_TO_RAD) )
#define m_asin(x) ( asin(x) * RAD_TO_DEG )
#define m_acos(x) ( acos(x) * RAD_TO_DEG )
#define m_atan(x) ( atan(x) * RAD_TO_DEG )
#define m_atan2(y,x) ( atan2(y,x) * RAD_TO_DEG )
#define m_square(x) ( (x)*(x) )
#define m_min(x,y) ( ((x)<(y))?(x):(y) )
#define m_max(x,y) ( ((x)>(y))?(x):(y) )
#define m_clamp(x,a,b) ( (int)x % (int)(b-a) + (int)a )

/*
 * Vector 3D.
 */
typedef float vec2_t[2]; 
typedef float vec3_t[3]; 
typedef float vec4_t[4]; 

/*
 * Make null vector.
 */
void zero_vec2(float *v);
void zero_vec3(float *v);
void zero_vec4(float *v);
#define v2_zero zero_vec2
#define v3_zero zero_vec3
#define v4_zero zero_vec4

/*
 * Set vector.
 */
void set_vec2(float *v, const float x, const float y);
void set_vec3(float *v, const float x, const float y, const float z);
void set_vec4(float *v, const float x, const float y, const float z, const float w);
#define v2_set set_vec2
#define v3_set set_vec3
#define v4_set set_vec4

/* 
 * Copy src in dst.
 */
void copy_vec2(float *dst, const float *src);
void copy_vec3(float *dst, const float *src);
void copy_vec4(float *dst, const float *src);
#define v2_copy copy_vec2
#define v3_copy copy_vec3
#define v4_copy copy_vec4

/* 
 * Addition : dst = a + b .
 */
void add_vec3(float *dst, const float *a, const float *b);
#define v3_add add_vec3

/*
 * Difference : dst = a - b .
 */
void sub_vec3(float *dst, const float *a, const float *b);
#define v3_sub sub_vec3
#define v3_diff v3_sub

/*
 * Minus.
 */
void negate_vec3(float *v);
#define v3_negate negate_vec3

/* 
 * Scale : dst = scale * src .
 */
void scale_vec3(float *v, const float div);
void scale2_vec3(float *dst, const float *src, const float scale);
void scale2_vec4(float *dst, const float *src, const float scale);
#define v3_scale scale_vec3
#define v3_scale2 scale2_vec3
#define v4_scale2 scale2_vec4

/*
 * Length.
 */
float length_vec3(const float *v);
#define v3_length length_vec3

/*
 * Normalize.
 */
void normalize_vec3(float *v);
void norm_vec3(float *v);
#define normalise_vec3 normalize_vec3
#define v3_normal normalize_vec3
#define v3_norm norm_vec3
#define v3_normalize v3_norm
#define v3_normalise v3_normalize

/*
 * Dot prod : a . b .
 */
float dot_vec3(const float *a, const float *b);
#define v3_dot dot_vec3
#define scalar_product_vec3 dot_vec3

/* 
 * Cross prod : cross = v1 ^ v2 .
 */
void cross_vec3(float *cross, const float *v1, const float *v2);
#define v3_cross cross_vec3
#define vector_product_vec3 cross_vec3

/*
 * Compute Euler angles from direction vector (with zero roll).
 */
void hpr_from_vec3(float *hpr, const float *dir);
#define v3_hrp hpr_from_vec3

/*
 * Linear intERPolation.
 */
void lerp_vec3(float *v, const float *from, const float *to, const float t);
#define v3_lerp lerp_vec3

/*
 * Distance between a and b.
 */
float dist_vec3(const float *a, const float *b);
#define v3_dist dist_vec3

/*
 * Test the libruary.
 */
void test_vectors(void);
#define v3_test test_vectors


#endif /* _VEC3D_H */
